import logging

# Base class for the tester.
from tests.base import Base

# Client context creator.
from xchembku_api.datafaces.context import Context as XchembkuDatafaceClientContext

# Object managing datafaces.
from xchembku_api.datafaces.datafaces import xchembku_datafaces_get_default
from xchembku_api.models.crystal_plate_model import CrystalPlateModel
from xchembku_api.models.crystal_well_autolocation_model import (
    CrystalWellAutolocationModel,
)
from xchembku_api.models.crystal_well_model import CrystalWellModel

# Server context creator.
from xchembku_lib.datafaces.context import Context as XchembkuDatafaceServerContext

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestCrystalWellAutolocationDirectSqlite:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_sqlite.yaml"
        CrystalWellAutolocationTester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class TestCrystalWellAutolocationDirectMysql:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_mysql.yaml"
        CrystalWellAutolocationTester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class TestCrystalWellAutolocationServiceSqlite:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_sqlite.yaml"
        CrystalWellAutolocationTester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class TestCrystalWellAutolocationServiceMysql:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_mysql.yaml"
        CrystalWellAutolocationTester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class CrystalWellAutolocationTester(Base):
    """
    Class to test the dataface autolocation-related endpoints.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """

        # Get the multiconf from the testing configuration yaml.
        multiconf = self.get_multiconf()

        # Load the multiconf into a dict.
        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the xchembku dataface.
        xchembku_dataface_specification = multiconf_dict[
            "xchembku_dataface_specification"
        ]

        # Make the server context.
        xchembku_server_context = XchembkuDatafaceServerContext(
            xchembku_dataface_specification
        )

        # Make the client context.
        xchembku_client_context = XchembkuDatafaceClientContext(
            xchembku_dataface_specification
        )

        # Start the xchembku server context which includes the direct or network-addressable service.
        async with xchembku_server_context:
            # Start the matching xchembku client context.
            async with xchembku_client_context:
                await self.__run_the_test(constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_the_test(self, constants, output_directory):
        """ """

        # Reference the dataface object which the context has set up as the default.
        dataface = xchembku_datafaces_get_default()

        # Make a plate for the wells we will create.
        crystal_plate_model = CrystalPlateModel(
            formulatrix__plate__id=1,
            barcode="xyzw",
            visit="cm00001-1",
        )

        # Write two well records.
        filename1 = "abc.jpg"
        crystal_well_model1 = CrystalWellModel(
            position="01A_1",
            crystal_plate_uuid=crystal_plate_model.uuid,
            filename=filename1,
        )
        filename2 = "xyz.jpg"
        crystal_well_model2 = CrystalWellModel(
            position="01A_2",
            crystal_plate_uuid=crystal_plate_model.uuid,
            filename=filename2,
        )
        await dataface.upsert_crystal_wells([crystal_well_model1, crystal_well_model2])

        # Fetch all the wells which need autolocation.
        crystal_well_models = await dataface.fetch_crystal_wells_needing_autolocation(
            limit=100
        )

        assert len(crystal_well_models) == 2

        assert crystal_well_models[0].filename == filename1
        assert crystal_well_models[1].filename == filename2

        # ----------------------------------------------------------------
        # Now try adding a crystal well autolocation.
        crystal_well_autolocation_model = CrystalWellAutolocationModel(
            crystal_well_uuid=crystal_well_model1.uuid
        )

        crystal_well_autolocation_model.number_of_crystals = 10
        await dataface.originate_crystal_well_autolocations(
            [crystal_well_autolocation_model]
        )

        # Fetch all the wells which need autolocation, which now there is only one.
        crystal_well_models = await dataface.fetch_crystal_wells_needing_autolocation(
            limit=100
        )

        # Now there is only one needing autolocation.
        assert len(crystal_well_models) == 1
        assert crystal_well_models[0].filename == filename2

        # ----------------------------------------------------------------
        # Now try adding an autolocation to the second well.
        crystal_well_autolocation_model = CrystalWellAutolocationModel(
            crystal_well_uuid=crystal_well_model2.uuid
        )

        crystal_well_autolocation_model.number_of_crystals = 10
        await dataface.originate_crystal_well_autolocations(
            [crystal_well_autolocation_model]
        )

        # Fetch all the wells which need autolocation.
        crystal_well_models = await dataface.fetch_crystal_wells_needing_autolocation(
            limit=100
        )

        # Now there are no more needing autolocation.
        assert len(crystal_well_models) == 0



import asyncio
import copy
import logging
from typing import Dict, List, Optional

from dls_normsql.constants import CommonFieldnames
from dls_utilpack.describe import describe

from xchembku_api.models.crystal_well_droplocation_model import (
    CrystalWellDroplocationModel,
)
from xchembku_lib.crystal_plate_objects.crystal_plate_objects import CrystalPlateObjects
from xchembku_lib.datafaces.direct_base import DirectBase

logger = logging.getLogger(__name__)


class DirectCrystalWellDroplocations(DirectBase):
    """ """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):

        # Lock allows only one coroutine to acquire it at a time.
        self.__upsert_lock = asyncio.Lock()

    # ----------------------------------------------------------------------------------------
    async def upsert_crystal_well_droplocations_serialized(
        self,
        records: List[Dict],
        only_fields: Optional[List[str]] = None,
        why=None,
    ) -> Dict:
        # We are being given json, so parse it into models.
        models = [CrystalWellDroplocationModel(**record) for record in records]
        # Return the method doing the work.
        return await self.upsert_crystal_well_droplocations(
            models,
            only_fields=only_fields,
            why=why,
        )

    # ----------------------------------------------------------------------------------------
    async def __add_confirmed_microns(self, model_dict: Dict, why=None) -> None:

        if why is not None:
            why = f"[CONFMIC] {why}"

        # Input model not updating confirmed target?
        if (
            "confirmed_target_x" not in model_dict
            or "confirmed_target_y" not in model_dict
        ):
            return

        from xchembku_api.models.crystal_well_filter_model import CrystalWellFilterModel

        filter = CrystalWellFilterModel(anchor=model_dict["crystal_well_uuid"])

        # Get the well record.
        crystal_well_models = await self.fetch_crystal_wells_needing_droplocation(
            filter,
            why=f"(crystal well for adding confirmed microns) {why}",
        )

        if len(crystal_well_models) == 0:
            raise RuntimeError(
                "database integrity error: no crystal well for droplocation upsert"
            )
        crystal_well_model = crystal_well_models[0]
        crystal_well_model.confirmed_target_x = model_dict["confirmed_target_x"]
        crystal_well_model.confirmed_target_y = model_dict["confirmed_target_y"]

        crystal_plate_object = CrystalPlateObjects().build_object(
            {"type": crystal_well_model.crystal_plate_thing_type}
        )

        x, y = crystal_plate_object.compute_drop_location_microns(
            crystal_well_model.dict()
        )

        logger.debug(f"{why} x is {x}, y is {y}")

        model_dict["confirmed_microns_x"] = x
        model_dict["confirmed_microns_y"] = y

    # ----------------------------------------------------------------------------------------
    async def upsert_crystal_well_droplocations(
        self,
        models: List[CrystalWellDroplocationModel],
        only_fields: Optional[List[str]] = None,
        why=None,
    ) -> Dict:
        """
        Caller provides the crystal well droplocation record with the fields to be updated.

        We don't insert for the same crystal_well_uuid twice.

        TODO: Find more efficient way to upsert_crystal_well_droplocations in batch.
        """

        if why is None:
            why = "upsert_crystal_well_droplocations"

        inserted_count = 0
        updated_count = 0

        # Loop over all the models to be upserted.
        for model in models:
            # Need to lock because of possible long time between query and eventual insert, user might double-click to send two droplocations quite quickly.
            # TODO: Reconsider direct_crystal_well_droplocations upsert logic to be tolerant of multiple processes possibly doing concurrent insert.
            # TODO: Add unit test for concurrent direct_crystal_well_droplocations upsert.
            async with self.__upsert_lock:
                model_dict = copy.deepcopy(model.dict())

                # Find any existing record for this model object.
                records = await self.query(
                    "SELECT * FROM crystal_well_droplocations WHERE crystal_well_uuid = ?",
                    subs=[model.crystal_well_uuid],
                    why=why,
                )

                if len(records) > 0:
                    logger.debug(
                        describe(
                            "crystal_well_droplocation record before update", records[0]
                        )
                    )
                    # Make a copy of the model record and remove some fields not to update.
                    model_dict.pop(CommonFieldnames.UUID)
                    model_dict.pop(CommonFieldnames.CREATED_ON)
                    if only_fields is not None:
                        for field in list(model_dict.keys()):
                            if field not in only_fields:
                                model_dict.pop(field)

                    # Convert confirmed target to microns and store in the dict.
                    await self.__add_confirmed_microns(model_dict, why=why)

                    # Don't update the crystal_well_uuid since it is used as the key.
                    if "crystal_well_uuid" in model_dict:
                        model_dict.pop("crystal_well_uuid")

                    result = await self.update(
                        "crystal_well_droplocations",
                        model_dict,
                        "(crystal_well_uuid = ?)",
                        subs=[model.crystal_well_uuid],
                        why=why,
                    )
                    updated_count += result.get("count", 0)

                    # Find any existing record for this model object.
                    records = await self.query(
                        "SELECT * FROM crystal_well_droplocations WHERE crystal_well_uuid = ?",
                        subs=[model.crystal_well_uuid],
                        why=why,
                    )
                    logger.debug(
                        describe(
                            "crystal_well_droplocation record after update", records[0]
                        )
                    )

                else:

                    # Convert confirmed target to microns and store in the dict.
                    await self.__add_confirmed_microns(model_dict, why=why)

                    await self.insert(
                        "crystal_well_droplocations",
                        [model_dict],
                        why=why,
                    )
                    inserted_count += 1

        return {
            "updated_count": updated_count,
            "inserted_count": inserted_count,
        }



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class CrystalWellFilterSortbyEnum(str, Enum):
    POSITION = "position"
    NUMBER_OF_CRYSTALS = "number_of_crystals"


class CrystalWellFilterModel(BaseModel):
    """
    Model containing crystal well query filter.
    """

    filename_pattern: Optional[str] = None
    visit: Optional[str] = None
    barcode: Optional[str] = None
    anchor: Optional[str] = None
    limit: Optional[int] = None
    sortby: Optional[CrystalWellFilterSortbyEnum] = None
    direction: Optional[int] = None
    is_decided: Optional[bool] = None
    is_usable: Optional[bool] = None
    is_exported_to_soakdb3: Optional[bool] = None



import logging

from xchembku_lib.datafaces.direct_base import DirectBase
from xchembku_lib.datafaces.direct_crystal_plates import DirectCrystalPlates
from xchembku_lib.datafaces.direct_crystal_well_autolocations import (
    DirectCrystalWellAutolocations,
)
from xchembku_lib.datafaces.direct_crystal_well_droplocations import (
    DirectCrystalWellDroplocations,
)
from xchembku_lib.datafaces.direct_crystal_wells import DirectCrystalWells
from xchembku_lib.datafaces.direct_soakdb3_crystal_wells import (
    DirectSoakdb3CrystalWells,
)

logger = logging.getLogger(__name__)


class Direct(
    DirectCrystalPlates,
    DirectCrystalWells,
    DirectCrystalWellAutolocations,
    DirectCrystalWellDroplocations,
    DirectSoakdb3CrystalWells,
    DirectBase,
):
    """ """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):
        DirectBase.__init__(self, specification)
        DirectCrystalWellDroplocations.__init__(self, specification)



import argparse
import asyncio

# Use standard logging in this module.
import logging

from dls_utilpack.require import require

# Base class for cli subcommands.
from xchembku_cli.subcommands.base import ArgKeywords, Base

# Servbase context creator.
from xchembku_lib.datafaces.context import Context as XchembkuDatafaceContext

logger = logging.getLogger()


# --------------------------------------------------------------
class Service(Base):
    """
    Start single service and block until ^C or remotely requested shutdown.
    """

    def __init__(self, args, mainiac):
        super().__init__(args)

    # ----------------------------------------------------------------------------------------
    def run(self):
        """ """

        # Run in asyncio event loop.
        asyncio.run(self.__run_coro())

    # ----------------------------------------------------------
    async def __run_coro(self):
        """
        Run the service as an asyncio coro.
        """

        # Load the configuration.
        multiconf_object = self.get_multiconf(vars(self._args))
        # Resolve the symbols and give configuration as a dict.
        multiconf_dict = await multiconf_object.load()

        # Get the specfication we want by keyword in the full configuration.
        specification = require(
            "configuration",
            multiconf_dict,
            "xchembku_dataface_specification",
        )

        # We need the context to always start the service as a coro.
        if "context" not in specification:
            specification["context"] = {}
        specification["context"]["start_as"] = "coro"

        # Make the xchembku service context from the specification in the configuration.
        context = XchembkuDatafaceContext(specification)

        # Open the xchembku context which starts the service process.
        async with context:
            # Wait for the coro to finish.
            await context.server.wait_for_shutdown()

    # ----------------------------------------------------------
    @staticmethod
    def add_arguments(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """
        Add arguments for this subcommand.

        This is a static method called from the main program.

        Args:
            parser (argparse.ArgumentParser): Parser object which has been created already.

        """

        parser.add_argument(
            "--configuration",
            "-c",
            help="Configuration file.",
            type=str,
            metavar="filename.yaml",
            default=None,
            dest=ArgKeywords.CONFIGURATION,
        )

        return parser



import logging
import os
import shutil

import pytest

# Formatting of testing log messages.
from dls_logformatter.dls_logformatter import DlsLogformatter

# Version of the package.
# from xchembku_lib.version import meta as version_meta

logger = logging.getLogger(__name__)


# --------------------------------------------------------------------------------
@pytest.fixture(scope="session")
def constants(request):

    constants = {}

    yield constants


# --------------------------------------------------------------------------------
@pytest.fixture(scope="session")
def logging_setup():
    # print("")

    formatter = DlsLogformatter(type="long")
    logger = logging.StreamHandler()
    logger.setFormatter(formatter)
    logging.getLogger().addHandler(logger)

    # Log level for all modules.
    logging.getLogger().setLevel("DEBUG")

    # Turn off noisy debug.
    logging.getLogger("asyncio").setLevel("WARNING")
    logging.getLogger("pika").setLevel("WARNING")
    logging.getLogger("stomp").setLevel("WARNING")
    logging.getLogger("luigi-interface").setLevel("WARNING")
    logging.getLogger("luigi.xchembku_scheduler").setLevel("INFO")
    logging.getLogger("urllib3.connectionpool").setLevel("INFO")

    logging.getLogger("xchembku_lib.things").setLevel("INFO")

    # Messages about starting and stopping services.
    logging.getLogger("xchembku_lib.base_aiohttp").setLevel("INFO")

    # All xchembku database sql commands.
    # logging.getLogger("dls_normsql.aiosqlite").setLevel("INFO")

    logging.getLogger("xchembku_lib.xchembku_contexts.classic").setLevel("INFO")
    logging.getLogger("xchembku_lib.xchembku_datafaces.context").setLevel("INFO")

    # Registering signal handler.
    logging.getLogger("dls_siggy_lib.signal").setLevel("INFO")

    # Cover the version.
    # logger.info("\n%s", (json.dumps(version_meta(), indent=4)))

    yield None


# --------------------------------------------------------------------------------
@pytest.fixture(scope="function")
def output_directory(request):
    # TODO: Better way to get a newline in conftest after pytest emits the test class name.
    print("")

    # Tmp directory which we can write into.
    output_directory = "/tmp/%s/%s/%s" % (
        "/".join(__file__.split("/")[-3:-1]),
        request.cls.__name__,
        request.function.__name__,
    )

    # Tmp directory which we can write into.
    if os.path.exists(output_directory):
        shutil.rmtree(output_directory, ignore_errors=False, onerror=None)
    os.makedirs(output_directory)

    # logger.debug("output_directory is %s" % (output_directory))

    yield output_directory



import asyncio
import logging
import os
import subprocess
import time
from typing import Optional

from dls_utilpack.describe import describe

# Base class for the tester.
from tests.base import Base
from xchembku_api.datafaces.context import Context as ClientContext

# Object managing datafaces.
from xchembku_api.models.crystal_plate_filter_model import CrystalPlateFilterModel
from xchembku_api.models.crystal_plate_model import CrystalPlateModel

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestCli2Sqlite:
    """
    Test that we can start the service (which uses sqlite) via the command line and talk to it.
    """

    def test(self, constants, logging_setup, output_directory):

        configuration_file = "tests/configurations/service_sqlite.yaml"
        Cli2Tester(configuration_file).main(
            constants,
            configuration_file,
            output_directory,
        )


# ----------------------------------------------------------------------------------------
class TestCli2Mysql:
    """
    Test that we can start the service (which uses mysql) via the command line and talk to it.
    """

    def test(self, constants, logging_setup, output_directory):

        configuration_file = "tests/configurations/service_mysql.yaml"
        Cli2Tester(configuration_file).main(
            constants,
            configuration_file,
            output_directory,
        )


# ----------------------------------------------------------------------------------------
class Cli2Tester(Base):
    """
    Class to test the dataface.
    """

    def __init__(self, configuration_file):
        Base.__init__(self)

        self.__configuration_file = configuration_file

    async def _main_coroutine(self, constants, output_directory):
        """ """

        # Command to run the service.
        xchembku_server_cli = [
            "python",
            "-m",
            "xchembku_cli.main",
            "service",
            "--verbose",
            "--configuration",
            self.__configuration_file,
        ]

        # Let the output_directory symbol be replaced in the multiconf.
        os.environ["output_directory"] = output_directory

        # Launch the service as a process.
        logger.debug(f"launching subprocess {' '.join(xchembku_server_cli)}")
        process = subprocess.Popen(
            xchembku_server_cli,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        try:
            # Read the same configuration which the service process reads.
            multiconf_object = self.get_multiconf()
            multiconf_dict = await multiconf_object.load()

            # Get a client context to the server in the process we just started.
            xchembku_specification = multiconf_dict["xchembku_dataface_specification"]
            xchembku_client1_context = ClientContext(xchembku_specification)
            async with xchembku_client1_context as dataface1:

                # Wait until process is able to give a non-exceptional health report.
                start_time = time.time()
                max_seconds = 5.0
                while True:
                    health = await dataface1.client_report_health()

                    # Check if health report contains an exception.
                    exception = health.get("exception")
                    if exception is None:
                        logger.debug(describe("health", health))
                        # Continue with test if no exception.
                        break

                    logger.debug(f"[CONNRETRY] retrying after {exception}")

                    if process.poll() is not None:
                        raise RuntimeError(
                            "server apprently died without being able to give a health check"
                        )

                    # Too much time has elapsed?
                    if time.time() - start_time > max_seconds:
                        raise RuntimeError(
                            f"server not answering within {max_seconds} seconds"
                        )

                    await asyncio.sleep(1.0)

                # Interact with the server now that it's up.
                visit = "cm00001-1"
                models = []
                models.append(
                    CrystalPlateModel(
                        formulatrix__plate__id=10,
                        barcode="xyz1",
                        visit=visit,
                    )
                )
                models.append(
                    CrystalPlateModel(
                        formulatrix__plate__id=20,
                        barcode="xyz2",
                        visit=visit,
                    )
                )
                models.append(
                    CrystalPlateModel(
                        formulatrix__plate__id=30,
                        barcode="xyz3",
                        visit=visit,
                    )
                )

                # Insert all three plates at once.
                await dataface1.upsert_crystal_plates(models)

                # Start a second concurrent client connection.
                xchembku_client2_context = ClientContext(xchembku_specification)
                async with xchembku_client2_context as dataface2:

                    # Check the filtered queries from the second connection.
                    # This verifies the previous upserts have been done.
                    await self.__check(
                        dataface2,
                        CrystalPlateFilterModel(),
                        3,
                        "all",
                    )

                    # Second client requests the shutdown.
                    await dataface2.client_shutdown()
        finally:
            try:
                # Wait for the process to finish and get the output.
                stdout_bytes, stderr_bytes = process.communicate(timeout=5)
            except subprocess.TimeoutExpired:
                # Timeout happens when client dies but server hasn't been told to shutdown.
                process.kill()
                stdout_bytes, stderr_bytes = process.communicate()

            # Get the return code of the process
            return_code = process.returncode
            logger.debug(f"server return_code is {return_code}")

            if len(stderr_bytes) > 0:
                logger.debug(
                    f"================================== server stderr is:\n{stderr_bytes.decode()}"
                )
            if len(stdout_bytes) > 0:
                logger.debug(
                    f"================================== server stdout is:\n{stdout_bytes.decode()}"
                )
            if len(stderr_bytes) > 0 or len(stdout_bytes) > 0:
                logger.debug("================================== end of server output")

        assert return_code == 0

    # ----------------------------------------------------------------------------------------

    async def __check(
        self,
        dataface,
        filter: CrystalPlateFilterModel,
        expected: int,
        note: str,
        formulatrix__plate__id: Optional[int] = None,
    ):
        """ """

        crystal_plate_models = await dataface.fetch_crystal_plates(filter)

        assert len(crystal_plate_models) == expected, note

        if formulatrix__plate__id is not None:
            assert (
                crystal_plate_models[0].formulatrix__plate__id == formulatrix__plate__id
            ), f"{note} formulatrix__plate__id"

        return crystal_plate_models



import logging
from typing import List

import pytest

# Base class for the tester.
from tests.base import Base

# Types which the CrystalPlateObjects factory can use to build an instance.
from xchembku_api.crystal_plate_objects.constants import (
    ThingTypes as CrystalPlateThingTypes,
)

# Client context creator.
from xchembku_api.datafaces.context import Context as XchembkuDatafaceClientContext

# Object managing datafaces.
from xchembku_api.datafaces.datafaces import xchembku_datafaces_get_default
from xchembku_api.models.crystal_plate_model import CrystalPlateModel
from xchembku_api.models.crystal_well_autolocation_model import (
    CrystalWellAutolocationModel,
)
from xchembku_api.models.crystal_well_droplocation_model import (
    CrystalWellDroplocationModel,
)
from xchembku_api.models.crystal_well_filter_model import (
    CrystalWellFilterModel,
    CrystalWellFilterSortbyEnum,
)
from xchembku_api.models.crystal_well_model import CrystalWellModel
from xchembku_lib.crystal_plate_objects.crystal_plate_objects import CrystalPlateObjects

# Server context creator.
from xchembku_lib.datafaces.context import Context as XchembkuDatafaceServerContext

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestCrystalWellDroplocation2DirectSqlite:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_sqlite.yaml"
        CrystalWellDroplocation2Tester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class TestCrystalWellDroplocation2DirectMysql:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_mysql.yaml"
        CrystalWellDroplocation2Tester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class TestCrystalWellDroplocation2ServiceSqlite:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_sqlite.yaml"
        CrystalWellDroplocation2Tester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class TestCrystalWellDroplocation2ServiceMysql:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_mysql.yaml"
        CrystalWellDroplocation2Tester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class CrystalWellDroplocation2Tester(Base):
    """
    Class to test the dataface droplocation-related endpoints.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """
        self.__injected_count = 0

        # Get the multiconf from the testing configuration yaml.
        multiconf = self.get_multiconf()

        # Load the multiconf into a dict.
        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the xchembku dataface.
        xchembku_dataface_specification = multiconf_dict[
            "xchembku_dataface_specification"
        ]

        # Make the server context.
        xchembku_server_context = XchembkuDatafaceServerContext(
            xchembku_dataface_specification
        )

        # Make the client context.
        xchembku_client_context = XchembkuDatafaceClientContext(
            xchembku_dataface_specification
        )

        # Start the xchembku server context which includes the direct or network-addressable service.
        async with xchembku_server_context:
            # Start the matching xchembku client context.
            async with xchembku_client_context:
                await self.__run_the_test(constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_the_test(self, constants, output_directory):
        """ """

        # Reference the dataface object which the context has set up as the default.
        dataface = xchembku_datafaces_get_default()

        # Make a plate for the wells we will create.
        self.__visit = "cm00001-1"
        self.__barcode = "xyzw"
        self.__crystal_plate_model = CrystalPlateModel(
            formulatrix__plate__id=1,
            barcode=self.__barcode,
            visit=self.__visit,
            thing_type=CrystalPlateThingTypes.SWISS3,
        )

        self.__crystal_plate_object = CrystalPlateObjects().build_object(
            {"type": CrystalPlateThingTypes.SWISS3}
        )

        # Write plate record.
        await dataface.upsert_crystal_plates(
            [self.__crystal_plate_model],
        )

        models = []

        # Inject some wells, all of which will belong to the plate we just made.
        models.append(await self.__inject(dataface, 1))  # 1
        models.append(await self.__inject(dataface, 2))  # 2
        models.append(await self.__inject(dataface, 3))  # 3
        models.append(await self.__inject(dataface, 3))  # 4
        models.append(await self.__inject(dataface, 2))  # 5
        models.append(await self.__inject(dataface, 1))  # 6

        # Check the filtered queries.
        await self.__check(
            dataface,
            CrystalWellFilterModel(
                sortby=CrystalWellFilterSortbyEnum.NUMBER_OF_CRYSTALS,
                direction=1,
            ),
            [3, 4, 2, 5, 1, 6],
            "all, forward",
        )

        await self.__check(
            dataface,
            CrystalWellFilterModel(
                sortby=CrystalWellFilterSortbyEnum.NUMBER_OF_CRYSTALS,
                direction=-1,
            ),
            [6, 1, 5, 2, 4, 3],
            "all, reverse",
        )

        await self.__check(
            dataface,
            CrystalWellFilterModel(
                anchor=models[1].uuid,
                sortby=CrystalWellFilterSortbyEnum.NUMBER_OF_CRYSTALS,
            ),
            [2],
            "anchored, single",
        )

        with pytest.raises(RuntimeError):
            # Does not work without a visit.
            await self.__check(
                dataface,
                CrystalWellFilterModel(
                    anchor=models[1].uuid,
                    sortby=CrystalWellFilterSortbyEnum.NUMBER_OF_CRYSTALS,
                    direction=-1,
                    limit=1,
                ),
                [],
                "anchored, no visit",
            )

        await self.__check(
            dataface,
            CrystalWellFilterModel(
                # Anchor at #2.
                anchor=models[1].uuid,
                visit=self.__visit,
                sortby=CrystalWellFilterSortbyEnum.NUMBER_OF_CRYSTALS,
                direction=1,
                limit=1,
                is_usable=True,
            ),
            [5],
            "anchored, forward",
        )

        await self.__check(
            dataface,
            CrystalWellFilterModel(
                # Anchor at #2.
                anchor=models[1].uuid,
                visit=self.__visit,
                sortby=CrystalWellFilterSortbyEnum.NUMBER_OF_CRYSTALS,
                direction=-1,
                limit=1,
            ),
            [4],
            "anchored, reverse, single",
        )

        await self.__check(
            dataface,
            CrystalWellFilterModel(
                # Anchor at #2.
                anchor=models[1].uuid,
                visit=self.__visit,
                sortby=CrystalWellFilterSortbyEnum.NUMBER_OF_CRYSTALS,
                direction=-1,
                limit=10,
                is_usable=True,
            ),
            [4, 3],
            "anchored, reverse, all",
        )

    # ----------------------------------------------------------------------------------------

    async def __check(
        self,
        dataface,
        filter: CrystalWellFilterModel,
        expected: List[int],
        note: str,
    ):
        """ """

        # Get the full crystal well with auto and confirmed drop locations.
        crystal_well_models = await dataface.fetch_crystal_wells_needing_droplocation(
            filter
        )

        for index, crystal_well_model in enumerate(crystal_well_models):
            logger.debug(
                f"{index+1}. position {crystal_well_model.position} has {crystal_well_model.number_of_crystals} crystals at uuid {crystal_well_model.uuid}"
            )

        expected_positions = []
        got_positions = []
        for index, crystal_well_model in enumerate(crystal_well_models):
            expected_positions.append("A%02da" % (expected[index]))
            got_positions.append(crystal_well_model.position)

        assert expected_positions == got_positions, note

    # ----------------------------------------------------------------------------------------

    async def __inject(self, dataface, number_of_crystals: int):
        """ """

        self.__injected_count += 1
        filename = "%02dA_1.jpg" % (self.__injected_count)
        position = "A%02da" % (self.__injected_count)

        # Create the well object.
        m = CrystalWellModel(
            position=position,
            crystal_plate_uuid=self.__crystal_plate_model.uuid,
            crystal_plate_thing_type=self.__crystal_plate_model.thing_type,
            filename=filename,
        )

        # Write well record.
        await dataface.upsert_crystal_wells([m])

        # Add a crystal well autolocation.
        ta = CrystalWellAutolocationModel(
            crystal_well_uuid=m.uuid,
            number_of_crystals=number_of_crystals,
            well_centroid_x=100,
            well_centroid_y=100,
        )
        await dataface.originate_crystal_well_autolocations([ta])

        # Add a crystal well droplocation.
        td = CrystalWellDroplocationModel(
            crystal_well_uuid=m.uuid,
            confirmed_target_x=150,
            confirmed_target_y=50,
            is_usable=True,
        )

        await dataface.upsert_crystal_well_droplocations([td])

        return m



from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple


class Interface(ABC):
    """
    This is the required API interface for an Authorizer object.
    """

    @abstractmethod
    # ----------------------------------------------------------------------------------------
    def get_well_count(self) -> int:
        pass

    # ----------------------------------------------------------------------------------------
    @abstractmethod
    def normalize_subwell_name(self, subwell_name: str) -> str:
        """
        Converts the name given to the subwell by Luigi into the "position" format shown in soakdb3.

        Args:
            subwell_name (str): stem part of the subwell's image filename

        Raises:
            ValueError: the value does not follow the convention

        Returns:
            str: the displayable subwell position
        """
        pass

    # ----------------------------------------------------------------------------------------
    @abstractmethod
    def compute_drop_location_microns(
        self,
        crystal_well_record: Dict,
    ) -> Tuple[Optional[int], Optional[int]]:
        pass



from typing import Optional

from pydantic import BaseModel


class CrystalPlateReportModel(BaseModel):
    """
    Model containing plate information.

    Typically this structure is populated and transmitted by the rockminer package.
    """

    uuid: str
    # ID from the Plate table.
    formulatrix__plate__id: int
    # Name from the formulatrix "experiment" tree node.
    formulatrix__experiment__name: Optional[str]
    # Directory stem where wells were mined from.
    # Also used by echolocator to format the export csv filename.
    rockminer_collected_stem: Optional[str] = None
    # The 4-letter barcode.
    barcode: str
    # The visit gleaned from the Formulatrix database.
    visit: str
    # A string which allows the CrytsalPlateObjects factory to make an instance.
    thing_type: str

    # TODO: Add proper pydantic date parsing/valiation to CREATED_ON fields.
    created_on: str

    # These are not in the database, but get returned from some queries.
    collected_count: int
    chimped_count: int
    undecided_count: int
    undecided_crystals_count: int
    decided_count: int
    decided_usable_count: int
    decided_unusable_count: int
    exported_count: int
    usable_unexported_count: int



import uuid
from typing import Optional

from pydantic import BaseModel


class CrystalWellModel(BaseModel):
    """
    Model containing well information.

    Typically this structure is populated and transmitted by the rockingester package.
    """

    uuid: str

    # This is the plate on which the well resides.
    crystal_plate_uuid: str

    # This is the well's position on the plate, such as A01_1, A01_2, etc.
    # TODO: Enforce crystal_well.position unique among wells in plate.
    position: str

    # This is the filename containing the well's image.
    filename: str

    # This is the size of the well's image.
    # Optional for now to avoid having to refactor all tests to include it on the constructur.
    # TODO: Make crystal_well_model width and height not optional.
    width: Optional[int] = None
    height: Optional[int] = None

    # This is the error reading and parsing the image file, if any.
    error: Optional[str]

    # TODO: Add proper pydantic date parsing/valiation to CREATED_ON fields.
    created_on: Optional[str] = None

    def __init__(self, **kwargs):
        """
        Constructor, takes keyword arguments which it assigns to model properties.
        """
        # Automatically cook up a uuid if it's not provided to the constructor.
        if "uuid" not in kwargs:
            kwargs["uuid"] = str(uuid.uuid4())
        super().__init__(**kwargs)



import logging
from typing import List, Optional

# Base class for the tester.
from tests.base import Base

# Types which the CrystalPlateObjects factory can use to build an instance.
from xchembku_api.crystal_plate_objects.constants import (
    ThingTypes as CrystalPlateThingTypes,
)

# Client context creator.
from xchembku_api.datafaces.context import Context as XchembkuDatafaceClientContext

# Object managing datafaces.
from xchembku_api.datafaces.datafaces import xchembku_datafaces_get_default
from xchembku_api.models.crystal_plate_filter_model import CrystalPlateFilterModel
from xchembku_api.models.crystal_plate_model import CrystalPlateModel
from xchembku_api.models.crystal_plate_report_model import CrystalPlateReportModel
from xchembku_api.models.crystal_well_autolocation_model import (
    CrystalWellAutolocationModel,
)
from xchembku_api.models.crystal_well_droplocation_model import (
    CrystalWellDroplocationModel,
)
from xchembku_api.models.crystal_well_model import CrystalWellModel

# Server context creator.
from xchembku_lib.datafaces.context import Context as XchembkuDatafaceServerContext

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestCrystalPlateReportDirectSqlite:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_sqlite.yaml"
        CrystalPlateReportTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestCrystalPlateReportDirectMysql:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_mysql.yaml"
        CrystalPlateReportTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestCrystalPlateReportServiceSqlite:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_sqlite.yaml"
        CrystalPlateReportTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestCrystalPlateReportServiceMysql:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_mysql.yaml"
        CrystalPlateReportTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class CrystalPlateReportTester(Base):
    """
    Class to test the dataface droplocation-related endpoints.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """
        self.__injected_count = 0

        # Get the multiconf from the testing configuration yaml.
        multiconf = self.get_multiconf()

        # Load the multiconf into a dict.
        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the xchembku dataface.
        xchembku_dataface_specification = multiconf_dict[
            "xchembku_dataface_specification"
        ]

        # Make the server context.
        xchembku_server_context = XchembkuDatafaceServerContext(
            xchembku_dataface_specification
        )

        # Make the client context.
        xchembku_client_context = XchembkuDatafaceClientContext(
            xchembku_dataface_specification
        )

        # Start the xchembku server context which includes the direct or network-addressable service.
        async with xchembku_server_context:
            # Start the matching xchembku client context.
            async with xchembku_client_context:
                await self.__run_the_test(constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_the_test(self, constants, output_directory):
        """ """

        # Reference the dataface object which the context has set up as the default.
        dataface = xchembku_datafaces_get_default()

        # Make a plate which will put one well, needing no intervention
        self.__visit = "cm00001-1"
        self.__barcode = "xyzv"
        self.__crystal_plate_model = CrystalPlateModel(
            formulatrix__plate__id=1,
            barcode=self.__barcode,
            visit=self.__visit,
            thing_type=CrystalPlateThingTypes.SWISS3,
        )

        # Write plate record.
        await dataface.upsert_crystal_plates(
            [self.__crystal_plate_model],
        )

        await self.__inject(
            dataface, True, True, True, is_exported_to_soakdb3=True
        )  # Decided usable, exported.

        # Make a plate for the wells we will create.
        self.__visit = "cm00001-1"
        self.__barcode = "xyzw"
        self.__crystal_plate_model = CrystalPlateModel(
            formulatrix__plate__id=2,
            barcode=self.__barcode,
            visit=self.__visit,
            thing_type=CrystalPlateThingTypes.SWISS3,
        )

        # Write plate record.
        await dataface.upsert_crystal_plates(
            [self.__crystal_plate_model],
        )
        # Inject some wells, all of which will belong to the plate we just made.
        await self.__inject(dataface, False, False)  # Not chimped.
        await self.__inject(dataface, False, False)  # Not chimped.

        await self.__inject(dataface, True, False)  # Not viewed, aka undecided.
        await self.__inject(  # No crystals, not viewed, aka undecided.
            dataface, True, False, number_of_crystals=0
        )

        await self.__inject(dataface, True, True)  # Viewed but undecided.
        await self.__inject(  # No crystals, viewed, but undecided.
            dataface, True, True, number_of_crystals=0
        )

        await self.__inject(dataface, True, True, False)  # Decided unusable.
        await self.__inject(dataface, True, True, False)  # Decided usable.
        await self.__inject(dataface, True, True, True)  # Decided usable.
        await self.__inject(dataface, True, True, True)  # Decided usable.
        await self.__inject(
            dataface, True, True, True, is_exported_to_soakdb3=True
        )  # Decided usable, exported.

        # Get the full list.
        crystal_plate_report_models: List[
            CrystalPlateReportModel
        ] = await dataface.report_crystal_plates(CrystalPlateFilterModel(direction=-1))
        assert len(crystal_plate_report_models) == 2

        crystal_plate_report_model: CrystalPlateReportModel = (
            crystal_plate_report_models[0]
        )
        assert crystal_plate_report_model.collected_count == 11
        assert crystal_plate_report_model.chimped_count == 9
        assert crystal_plate_report_model.undecided_count == 4
        assert crystal_plate_report_model.undecided_crystals_count == 2
        assert crystal_plate_report_model.decided_count == 5
        assert crystal_plate_report_model.decided_usable_count == 3
        assert crystal_plate_report_model.decided_unusable_count == 2
        assert crystal_plate_report_model.exported_count == 1
        assert crystal_plate_report_model.usable_unexported_count == 2

        # ----------------------------------------------------------------------
        # Get those which have no undecided crystals and no unexported wells, aka need no intervention.
        crystal_plate_report_models: List[
            CrystalPlateReportModel
        ] = await dataface.report_crystal_plates(
            CrystalPlateFilterModel(needing_intervention=False)
        )

        assert len(crystal_plate_report_models) == 1
        assert crystal_plate_report_models[0].formulatrix__plate__id == 1

        # ----------------------------------------------------------------------
        # Get those which have either undecided crystals or unexported wells, aka need intervention.
        crystal_plate_report_models: List[
            CrystalPlateReportModel
        ] = await dataface.report_crystal_plates(
            CrystalPlateFilterModel(needing_intervention=True)
        )

        assert len(crystal_plate_report_models) == 1
        assert crystal_plate_report_models[0].formulatrix__plate__id == 2

    # ----------------------------------------------------------------------------------------

    async def __inject(
        self,
        dataface,
        autolocation: bool,
        droplocation: bool,
        is_usable: Optional[bool] = None,
        number_of_crystals: Optional[int] = None,
        is_exported_to_soakdb3: Optional[bool] = None,
    ):
        """ """

        letter = "A"
        if self.__injected_count > 3:
            letter = "B"

        self.__injected_count += 1
        filename = "%02d%s_1.jpg" % (self.__injected_count, letter)
        position = "%s%02da" % (letter, self.__injected_count)

        # Create the well object.
        m = CrystalWellModel(
            position=position,
            crystal_plate_uuid=self.__crystal_plate_model.uuid,
            crystal_plate_thing_type=self.__crystal_plate_model.thing_type,
            filename=filename,
        )

        # Write well record.
        await dataface.upsert_crystal_wells([m])

        if autolocation:
            if number_of_crystals is None:
                number_of_crystals = self.__injected_count
            # Add a crystal well autolocation.
            ta = CrystalWellAutolocationModel(
                crystal_well_uuid=m.uuid,
                number_of_crystals=number_of_crystals,
                well_centroid_x=100,
                well_centroid_y=100,
            )

            await dataface.originate_crystal_well_autolocations([ta])

        if droplocation:
            # Add a crystal well droplocation.
            td = CrystalWellDroplocationModel(
                crystal_well_uuid=m.uuid,
                confirmed_target_x=150,
                confirmed_target_y=50,
                is_usable=is_usable,
            )
            if is_exported_to_soakdb3 is not None:
                td.is_exported_to_soakdb3 = is_exported_to_soakdb3

            await dataface.upsert_crystal_well_droplocations([td])

        return m



import logging

import pytest

from xchembku_lib.crystal_plate_objects.swiss3 import Swiss3

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestSwiss3:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        swiss3 = Swiss3({})

        position = swiss3.normalize_subwell_name("98ju_01A_1")
        assert position == "A01a"
        position = swiss3.normalize_subwell_name("98ju_08H_2")
        assert position == "H08c"
        position = swiss3.normalize_subwell_name("98ju_12H_3")
        assert position == "H12d"

        # Check invalid name formats.
        with pytest.raises(ValueError) as excinfo:
            swiss3.normalize_subwell_name("98ab_273A_1")
        assert "expected format" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            swiss3.normalize_subwell_name("98ju_01H_1.jpg")
        assert "expected format" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            swiss3.normalize_subwell_name("98ju_01J_1")
        assert "expected format" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            swiss3.normalize_subwell_name("98ju_01A_4")
        assert "expected format" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            swiss3.normalize_subwell_name("98j_01A_1")
        assert "expected format" in str(excinfo.value)



from typing import Optional

from pydantic import BaseModel


class CrystalPlateFilterModel(BaseModel):
    """
    Model containing crystal plate query filter.
    """

    uuid: Optional[str] = None
    visit: Optional[str] = None
    barcode: Optional[str] = None
    include_errors: Optional[bool] = None
    limit: Optional[int] = None
    direction: Optional[int] = None

    # Allow searching starting from a particular plate id.
    from_formulatrix__plate__id: Optional[int] = None

    # These choices are only used in the report query.
    needing_intervention: Optional[bool] = None



import logging

import dls_mainiac_lib.version
import dls_normsql.version
import dls_servbase_lib.version
import dls_utilpack.version

import xchembku_lib

logger = logging.getLogger(__name__)


# ----------------------------------------------------------
def version():
    """
    Current version.
    """

    return xchembku_lib.__version__


# ----------------------------------------------------------
def meta(given_meta=None):
    """
    Returns version information as a dict.
    Adds version information to given meta, if any.
    """
    s = {}
    s["xchembku_lib"] = version()

    s.update(dls_servbase_lib.version.meta())
    s.update(dls_utilpack.version.meta())
    s.update(dls_mainiac_lib.version.meta())
    s.update(dls_normsql.version.meta())

    try:
        import setproctitle

        setproctitle.__version__
        s["setproctitle"] = setproctitle.__version__
    except Exception:
        s["setproctitle"] = "unavailable"

    if given_meta is not None:
        given_meta.update(s)
    else:
        given_meta = s
    return given_meta



class ThingTypes:
    SWISS2 = "xchembku_api::crystal_plate_objects::swiss2"
    SWISS3 = "xchembku_api::crystal_plate_objects::swiss3"
    MITEGEN = "xchembku_api::crystal_plate_objects::mitegen"


# Names that appear in the Formulatrix database, and how they map to our class types.
TREENODE_NAMES_TO_THING_TYPES = {
    "SWISSci_3Drop": ThingTypes.SWISS3,
    "SWISSci_3drop": ThingTypes.SWISS3,
    # When adding these plate types, need to add corresponding entries to factory in crystal_plate_objects.py.
    # TODO: Add 2drop and mitegen plate types.
    # "SWISSci_2drop": ThingTypes.SWISS2,
    # "Mitegen_insitu1": ThingTypes.MITEGEN,
    # "MiTInSitu": ThingTypes.MITEGEN,
}



# Use standard logging in this module.
import logging

# Class managing list of things.
from dls_utilpack.things import Things

# Types.
from xchembku_api.datafaces.constants import Types

# Exceptions.
from xchembku_api.exceptions import NotFound

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------------------


class Datafaces(Things):
    """
    List of available xchembku_datafaces.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, name=None):
        Things.__init__(self, name)

    # ----------------------------------------------------------------------------------------
    def build_object(self, specification):
        """"""

        xchembku_dataface_class = self.lookup_class(specification["type"])

        try:
            xchembku_dataface_object = xchembku_dataface_class(specification)
        except Exception as exception:
            raise RuntimeError(
                "unable to build xchembku_dataface object for type %s"
                % (xchembku_dataface_class)
            ) from exception

        return xchembku_dataface_object

    # ----------------------------------------------------------------------------------------
    def lookup_class(self, class_type):
        """"""

        if class_type == Types.AIOHTTP:
            from xchembku_lib.datafaces.aiohttp import Aiohttp

            return Aiohttp

        elif class_type == Types.DIRECT:
            from xchembku_lib.datafaces.direct import Direct

            return Direct

        raise NotFound(
            "unable to get xchembku_dataface class for type %s" % (class_type)
        )



import uuid
from typing import Optional

from pydantic import BaseModel


class CrystalPlateModel(BaseModel):
    """
    Model containing plate information.

    Typically this structure is populated and transmitted by the rockingester package.
    """

    uuid: str
    # ID from the Plate table.
    formulatrix__plate__id: Optional[int]
    # Name from the formulatrix "experiment" tree node.
    formulatrix__experiment__name: Optional[str]
    # Directory stem where wells were mined from.
    # Also used by echolocator to format the export csv filename.
    rockminer_collected_stem: Optional[str] = None
    # The 4-letter barcode.
    barcode: str
    # The visit gleaned from the Formulatrix database.
    visit: Optional[str]
    # A string which allows the CrytsalPlateObjects factory to make an instance.
    thing_type: Optional[str] = None
    # Error string.
    error: Optional[str] = None

    # TODO: Add proper pydantic date parsing/valiation to CREATED_ON fields.
    created_on: Optional[str] = None

    def __init__(self, **kwargs):
        # Automatically cook up a uuid if it's not provided to the constructor.
        if "uuid" not in kwargs:
            kwargs["uuid"] = str(uuid.uuid4())
        super().__init__(**kwargs)



import logging

# Fieldnames common to all databases.
from dls_normsql.constants import CommonFieldnames

# Base class for table definitions.
from dls_normsql.table_definition import TableDefinition

from xchembku_api.models.crystal_plate_model import CrystalPlateModel
from xchembku_api.models.crystal_well_autolocation_model import (
    CrystalWellAutolocationModel,
)
from xchembku_api.models.crystal_well_droplocation_model import (
    CrystalWellDroplocationModel,
)
from xchembku_api.models.crystal_well_model import CrystalWellModel

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class CrystalPlatesTable(TableDefinition):
    # ----------------------------------------------------------------------------------------
    def __init__(self):
        model_class = CrystalPlateModel
        table_name = "crystal_plates"

        TableDefinition.__init__(self, table_name)

        fields = model_class.__fields__
        for field_name, field in fields.items():
            field_type = field.type_

            if field_name == "uuid":
                # All images have a unique autoid field.
                self.fields[CommonFieldnames.UUID] = {
                    "type": "TEXT PRIMARY KEY",
                    "index": True,
                }

            else:
                if field_type == int:
                    sql_type = "INTEGER"
                elif field_type == str:
                    sql_type = "TEXT"
                elif field_type == float:
                    sql_type = "REAL"
                elif field_type == bool:
                    sql_type = "BOOLEAN"

                self.fields[field_name] = {"type": sql_type}

        # Add indexes.
        self.fields["formulatrix__plate__id"]["index"] = True
        self.fields["formulatrix__experiment__name"]["index"] = True
        self.fields["barcode"]["index"] = True
        self.fields["visit"]["index"] = True
        self.fields["thing_type"]["index"] = True
        self.fields[CommonFieldnames.CREATED_ON]["index"] = True


# ----------------------------------------------------------------------------------------
class CrystalWellsTable(TableDefinition):
    # ----------------------------------------------------------------------------------------
    def __init__(self):
        model_class = CrystalWellModel
        table_name = "crystal_wells"

        TableDefinition.__init__(self, table_name)

        fields = model_class.__fields__
        for field_name, field in fields.items():
            field_type = field.type_

            if field_name == "uuid":
                # All images have a unique autoid field.
                self.fields[CommonFieldnames.UUID] = {
                    "type": "TEXT PRIMARY KEY",
                    "index": True,
                }

            else:
                if field_type == int:
                    sql_type = "INTEGER"
                elif field_type == str:
                    sql_type = "TEXT"
                elif field_type == float:
                    sql_type = "REAL"
                elif field_type == bool:
                    sql_type = "BOOLEAN"

                self.fields[field_name] = {"type": sql_type}

        # Add indexes.
        self.fields["position"]["index"] = True
        self.fields["filename"]["index"] = True
        self.fields["crystal_plate_uuid"]["index"] = True
        self.fields[CommonFieldnames.CREATED_ON]["index"] = True


# ----------------------------------------------------------------------------------------
class CrystalWellAutolocationsTable(TableDefinition):
    # ----------------------------------------------------------------------------------------
    def __init__(self):
        model_class = CrystalWellAutolocationModel
        table_name = "crystal_well_autolocations"

        TableDefinition.__init__(self, table_name)

        fields = model_class.__fields__
        for field_name, field in fields.items():
            field_type = field.type_

            if field_name == "uuid":
                # All images have a unique autoid field.
                self.fields[CommonFieldnames.UUID] = {
                    "type": "TEXT PRIMARY KEY",
                    "index": True,
                }

            else:
                if field_type == int:
                    sql_type = "INTEGER"
                elif field_type == str:
                    sql_type = "TEXT"
                elif field_type == float:
                    sql_type = "REAL"
                elif field_type == bool:
                    sql_type = "BOOLEAN"

                self.fields[field_name] = {"type": sql_type}

        # Remove attribute of the model that doesn't get stored in the database.
        self.fields.pop("crystal_coordinates")

        # Add indexes.
        self.fields["crystal_well_uuid"]["index"] = True
        self.fields["number_of_crystals"]["index"] = True
        self.fields[CommonFieldnames.CREATED_ON]["index"] = True


# ----------------------------------------------------------------------------------------
class CrystalWellDroplocationsTable(TableDefinition):
    # ----------------------------------------------------------------------------------------
    def __init__(self):
        model_class = CrystalWellDroplocationModel
        table_name = "crystal_well_droplocations"

        TableDefinition.__init__(self, table_name)

        fields = model_class.__fields__
        for field_name, field in fields.items():
            field_type = field.type_

            if field_name == "uuid":
                # All images have a unique dropid field.
                self.fields[CommonFieldnames.UUID] = {
                    "type": "TEXT PRIMARY KEY",
                    "index": True,
                }

            else:
                if field_type == int:
                    sql_type = "INTEGER"
                elif field_type == str:
                    sql_type = "TEXT"
                elif field_type == float:
                    sql_type = "REAL"
                elif field_type == bool:
                    sql_type = "BOOLEAN"

                self.fields[field_name] = {"type": sql_type}

        # Add indexes.
        self.fields["crystal_well_uuid"]["index"] = True
        self.fields["is_usable"]["index"] = True
        self.fields["is_exported_to_soakdb3"]["index"] = True
        self.fields[CommonFieldnames.CREATED_ON]["index"] = True



import logging

# All the tables.
from xchembku_api.databases.table_definitions import (
    CrystalPlatesTable,
    CrystalWellAutolocationsTable,
    CrystalWellDroplocationsTable,
    CrystalWellsTable,
)

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class DatabaseDefinition:
    """
    Class which defines the database tables and revision migration path.
    Used in concert with the normsql class.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self):
        """
        Construct object.  Do not connect to database.
        """

        self.LATEST_REVISION = 5

    # ----------------------------------------------------------------------------------------
    async def apply_revision(self, database, revision):

        logger.debug(f"applying revision {revision}")

        if revision == 3:

            # Add crytal plate formulatrix__experiment__name field and index.
            await database.execute(
                "ALTER TABLE crystal_plates ADD COLUMN formulatrix__experiment__name TEXT",
                why=f"revision {revision}: new column",
            )
            await database.execute(
                "CREATE INDEX %s_%s ON %s(%s)"
                % (
                    "crystal_plates",
                    "formulatrix__experiment__name",
                    "crystal_plates",
                    "formulatrix__experiment__name",
                )
            )

        if revision == 4:
            # Add crytal plate formulatrix__experiment__name field and index.
            await database.execute(
                "ALTER TABLE crystal_well_droplocations ADD COLUMN is_exported_to_soakdb3 BOOLEAN",
                why="revision {revision}: new column",
            )
            await database.execute(
                "CREATE INDEX %s_%s ON %s(%s)"
                % (
                    "crystal_well_droplocations",
                    "is_exported_to_soakdb3",
                    "crystal_well_droplocations",
                    "is_exported_to_soakdb3",
                )
            )

        if revision == 5:
            # Add crytal plate error field.
            await database.execute(
                "ALTER TABLE crystal_plates ADD COLUMN error TEXT",
                why="revision {revision}: new column",
            )

    # ----------------------------------------------------------------------------------------
    async def add_table_definitions(self, database):
        """
        Make all the table definitions.
        """

        # Table schemas in our database.
        database.add_table_definition(CrystalPlatesTable())
        database.add_table_definition(CrystalWellsTable())
        database.add_table_definition(CrystalWellAutolocationsTable())
        database.add_table_definition(CrystalWellDroplocationsTable())



import copy
import logging
from typing import Dict, List, Union

from dls_normsql.constants import CommonFieldnames

from xchembku_api.models.crystal_plate_filter_model import CrystalPlateFilterModel
from xchembku_api.models.crystal_plate_model import CrystalPlateModel
from xchembku_api.models.crystal_plate_report_model import CrystalPlateReportModel
from xchembku_lib.datafaces.direct_base import DirectBase

logger = logging.getLogger(__name__)


class DirectCrystalPlates(DirectBase):
    """ """

    # ----------------------------------------------------------------------------------------
    async def upsert_crystal_plates_serialized(
        self,
        records: List[Dict],
        why=None,
    ) -> Dict:
        # We are being given json, so parse it into models.
        models = [CrystalPlateModel(**record) for record in records]
        # Return the method doing the work.
        return await self.upsert_crystal_plates(models, why=why)

    # ----------------------------------------------------------------------------------------
    async def upsert_crystal_plates(
        self,
        models: List[CrystalPlateModel],
        why="upsert_crystal_plates",
    ) -> Dict:
        """
        Caller provides the crystal plate record with the fields to be updated.

        We don't insert the same formulatrix__plate__id twice.

        TODO: Find more efficient way to upsert_crystal_plates in batch.
        """

        inserted_count = 0
        updated_count = 0

        # Loop over all the models to be upserted.
        for model in models:
            # Find any existing record for this model object.
            records = await self.query(
                "SELECT * FROM crystal_plates WHERE formulatrix__plate__id = ?",
                subs=[model.formulatrix__plate__id],
                why=why,
            )

            if len(records) > 0:
                # Make a copy of the model record and remove some fields not to update.
                model_copy = copy.deepcopy(model.dict())
                model_copy.pop(CommonFieldnames.UUID)
                model_copy.pop(CommonFieldnames.CREATED_ON)
                model_copy.pop("formulatrix__plate__id")
                result = await self.update(
                    "crystal_plates",
                    model_copy,
                    "(formulatrix__plate__id = ?)",
                    subs=[model.formulatrix__plate__id],
                    why=why,
                )
                updated_count += result.get("count", 0)
            else:
                await self.insert(
                    "crystal_plates",
                    [model.dict()],
                    why=why,
                )
                inserted_count += 1

        return {
            "updated_count": updated_count,
            "inserted_count": inserted_count,
        }

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_plates_serialized(
        self, filter: Dict, why=None
    ) -> List[Dict]:
        """
        Caller provides the filters for selecting which crystal plates.
        Returns records from the database.
        """

        # Get the models from the direct call.
        models = await self.fetch_crystal_plates(
            CrystalPlateFilterModel(**filter), why=why
        )

        # Serialize models into dicts to give to the response.
        records = [model.dict() for model in models]

        return records

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_plates(
        self, filter: CrystalPlateFilterModel, why=None
    ) -> List[CrystalPlateModel]:
        """
        Plates need a droplocation if they have an autolocation but no droplocation.
        """

        if why is None:
            why = "API fetch_crystal_plates"

        # Build the individual pieces of the SQL query.
        subs: List[Union[str, int]] = []
        orderby = self.__build_orderby(filter)
        where = self.__build_where(filter, subs)
        fields = self.__build_fields(filter)
        joins = self.__build_joins(filter)

        # Glue them together.
        main_query = "\nSELECT" + fields + joins + where + "\n" + orderby

        if filter.limit is not None:
            main_query += f"\nLIMIT {filter.limit}"

        records = await self.query(main_query, subs=subs, why=why)

        # Parse the records returned by sql into models.
        models = [CrystalPlateModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def report_crystal_plates_serialized(
        self, filter: Dict, why=None
    ) -> List[Dict]:
        """
        Caller provides the filters for selecting which crystal plates.
        Returns records from the database.
        """

        # Get the models from the direct call.
        models = await self.report_crystal_plates(
            CrystalPlateFilterModel(**filter), why=why
        )

        # Serialize models into dicts to give to the response.
        records = [model.dict() for model in models]

        return records

    # ----------------------------------------------------------------------------------------
    async def report_crystal_plates(
        self, filter: CrystalPlateFilterModel, why=None
    ) -> List[CrystalPlateReportModel]:
        """
        Plates need a droplocation if they have an autolocation but no droplocation.
        """

        if why is None:
            why = "API report_crystal_plates"

        # Build the individual pieces of the SQL query.
        subs: List[Union[str, int]] = []
        orderby = self.__build_orderby(filter, is_for_report=True)
        where = self.__build_where(filter, subs, is_for_report=True)
        fields = self.__build_fields(filter, is_for_report=True)
        joins = self.__build_joins(filter, is_for_report=True)

        # Glue them together.
        main_query = "\nSELECT" + fields + joins + where + "\n" + orderby

        if filter.limit is not None:
            main_query += f"\nLIMIT {filter.limit}"

        records = await self.query(main_query, subs=subs, why=why)

        # Parse the records returned by sql into models.
        models = [CrystalPlateReportModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    def __build_fields(
        self,
        filter: CrystalPlateFilterModel,
        is_for_report: bool = False,
    ) -> str:
        """
        Wells need a droplocation if they have an autolocation.
        """

        fields = ["crystal_plates.*"]

        self.__usable_unexported_count = (
            "(COALESCE(decided_usable.count, 0) - COALESCE(exported.count, 0))"
        )
        self.__undecided_crystals_count = "COALESCE(undecided_crystals.count, 0)"

        if is_for_report:
            fields.append("COALESCE(collected.count, 0) AS collected_count")
            fields.append("COALESCE(chimped.count, 0) AS chimped_count")
            fields.append(
                "COALESCE(chimped.count, 0) - COALESCE(decided.count, 0) AS undecided_count"
            )
            fields.append("COALESCE(decided.count, 0) AS decided_count")
            fields.append("COALESCE(decided_usable.count, 0) AS decided_usable_count")
            fields.append(
                "COALESCE(decided.count, 0) - COALESCE(decided_usable.count, 0) AS decided_unusable_count"
            )
            fields.append("COALESCE(exported.count, 0) AS exported_count")
            fields.append(
                f"{self.__usable_unexported_count} AS usable_unexported_count"
            )
            fields.append(
                f"{self.__undecided_crystals_count} AS undecided_crystals_count"
            )

        return "\n  " + ",\n  ".join(fields)

    # ----------------------------------------------------------------------------------------
    def __build_joins(
        self,
        filter: CrystalPlateFilterModel,
        is_for_report: bool = False,
    ) -> str:
        """
        Wells need a droplocation if they have an autolocation.
        """

        joins = ["crystal_plates"]

        if is_for_report:
            all = "SELECT crystal_plate_uuid, COUNT(*) AS count FROM crystal_wells"
            chimped = f"{all} JOIN crystal_well_autolocations ON crystal_well_autolocations.crystal_well_uuid = crystal_wells.uuid"
            viewed = f"{all} JOIN crystal_well_droplocations ON crystal_well_droplocations.crystal_well_uuid = crystal_wells.uuid"
            both = (
                f"{all}"
                "\n  JOIN crystal_well_autolocations ON crystal_well_autolocations.crystal_well_uuid = crystal_wells.uuid"
                "\n  LEFT JOIN crystal_well_droplocations ON crystal_well_droplocations.crystal_well_uuid = crystal_wells.uuid"
                "\n  WHERE (number_of_crystals > 0) AND (is_usable is NULL) "
            )
            joins.append(
                f"LEFT JOIN ({all} GROUP BY crystal_plate_uuid) AS collected"
                f"\n    ON collected.crystal_plate_uuid = crystal_plates.uuid"
            )
            joins.append(
                f"LEFT JOIN ({chimped} GROUP BY crystal_plate_uuid) AS chimped"
                f"\n    ON chimped.crystal_plate_uuid = crystal_plates.uuid"
            )
            joins.append(
                f"LEFT JOIN ({viewed} WHERE (is_usable IS NOT NULL) GROUP BY crystal_plate_uuid) AS decided"
                f"\n    ON decided.crystal_plate_uuid = crystal_plates.uuid"
            )
            joins.append(
                f"LEFT JOIN ({viewed} WHERE (is_usable = True) GROUP BY crystal_plate_uuid) AS decided_usable"
                f"\n    ON decided_usable.crystal_plate_uuid = crystal_plates.uuid"
            )
            joins.append(
                f"LEFT JOIN ({viewed} WHERE (is_exported_to_soakdb3 = True) GROUP BY crystal_plate_uuid) AS exported"
                f"\n    ON exported.crystal_plate_uuid = crystal_plates.uuid"
            )
            joins.append(
                f"LEFT JOIN ({both} GROUP BY crystal_plate_uuid) AS undecided_crystals"
                f"\n    ON undecided_crystals.crystal_plate_uuid = crystal_plates.uuid"
            )

        return "\nFROM " + "\n  ".join(joins)

    # ----------------------------------------------------------------------------------------
    def __build_where(
        self,
        filter: CrystalPlateFilterModel,
        subs: List[Union[str, int]],
        is_for_report: bool = False,
    ) -> str:
        """
        Wells need a droplocation if they have an autolocation.
        """

        where = "WHERE"
        sql = ""

        if filter.uuid is not None:
            sql += f"\n{where} uuid = ?"
            subs.append(filter.uuid)
            where = "AND"

        if filter.visit is not None:
            sql += f"\n{where} visit = ?"
            subs.append(filter.visit)
            where = "AND"

        if filter.barcode is not None:
            sql += f"\n{where} barcode = ?"
            subs.append(filter.barcode)
            where = "AND"

        if filter.barcode is None:
            # Default, if not specified, is to exclude plates with errors.
            if filter.include_errors is None or filter.include_errors is False:
                sql += f"\n{where} error IS NULL"
                where = "AND"

        if filter.from_formulatrix__plate__id is not None:
            if filter.direction == -1:
                sql += f"\n{where} formulatrix__plate__id < ?"
            else:
                sql += f"\n{where} formulatrix__plate__id > ?"
            subs.append(filter.from_formulatrix__plate__id)
            where = "AND"

        if filter.needing_intervention is not None:
            if filter.needing_intervention is True:
                sql += "\n/* Those needing intervention. */"
                sql += f"\n{where} ({self.__undecided_crystals_count} > 0 OR {self.__usable_unexported_count} > 0)"
            else:
                sql += "\n/* Those NOT needing intervention. */"
                sql += f"\n{where} ({self.__undecided_crystals_count} = 0 AND {self.__usable_unexported_count} = 0)"

        return sql

    # ----------------------------------------------------------------------------------------
    def __build_orderby(
        self,
        filter: CrystalPlateFilterModel,
        is_for_report: bool = False,
    ) -> str:

        sql = ""

        sql_direction = "ASC"
        if filter.direction == -1:
            sql_direction = "DESC"

        order_by = f"crystal_plates.formulatrix__plate__id {sql_direction}"

        sql += f"ORDER BY {order_by}"

        return sql



# Use standard logging in this module.
import logging
from typing import List

# Class managing list of things.
from dls_utilpack.things import Things

# Types which the CrystalPlateObjects factory can use to build an instance.
from xchembku_api.crystal_plate_objects.constants import ThingTypes

# Interface fulfilled by all CrystalPlateObject isntances.
from xchembku_api.crystal_plate_objects.interface import (
    Interface as CrystalPlateInterface,
)

# Exceptions.
from xchembku_api.exceptions import NotFound

logger = logging.getLogger(__name__)


class CrystalPlateObjects(Things):
    """
    List of available crystal plate object types.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, name=None):
        Things.__init__(self, name)

    # ----------------------------------------------------------------------------------------
    def get_treenode_names(self) -> List[str]:
        """
        Returns list of all the Formulatrix treenode names which are interesting to soakdb3.

        Returns:
            List[str]: names
        """

        return list(self.__treenode_names_to_thing_type.keys())

    # ----------------------------------------------------------------------------------------
    def build_object(self, specification) -> CrystalPlateInterface:
        """ """

        object_class = self.lookup_class(specification["type"])

        try:
            object_instance = object_class(specification)
        except Exception as exception:
            raise RuntimeError(
                "unable to build object for type %s" % object_class
            ) from exception

        return object_instance

    # ----------------------------------------------------------------------------------------
    def lookup_class(self, class_type):
        """
        Given the class type as string, return a class object.
        """

        if class_type == ThingTypes.SWISS3:
            from xchembku_lib.crystal_plate_objects.swiss3 import Swiss3

            return Swiss3

        raise NotFound("unable to get class for type %s" % (class_type))



import logging
from typing import Dict, List, Optional

from dls_utilpack.callsign import callsign
from dls_utilpack.describe import describe
from dls_utilpack.require import require

# Soakdb3 dataface/database.
from soakdb3_api.databases.constants import Tablenames
from soakdb3_api.datafaces.datafaces import Datafaces as Soakdb3ApiDatafaces
from soakdb3_api.models.crystal_well_model import (
    CrystalWellModel as Soakdb3CrystalWellModel,
)

from xchembku_lib.datafaces.direct_base import DirectBase

logger = logging.getLogger(__name__)


class DirectSoakdb3CrystalWells(DirectBase):
    """ """

    # ----------------------------------------------------------------------------------------
    async def inject_soakdb3_crystal_wells_serialized(
        self,
        visitid: str,
        records: List[Dict],
        why: Optional[str] = None,
    ) -> Dict:
        # We are being given json, so parse it into models.
        models = [Soakdb3CrystalWellModel(**record) for record in records]
        # Return the method doing the work.
        return await self.inject_soakdb3_crystal_wells(visitid, models, why=why)

    # ----------------------------------------------------------------------------------------
    async def disconnect_soakdb3_crystal_wells_mixin(self):
        """
        Called from base class disconnect.
        """

        if (
            hasattr(self, "soakdb3_dataface_client")
            and self.soakdb3_dataface_client is not None
        ):
            await self.soakdb3_dataface_client.close_client_session()

    # ----------------------------------------------------------------------------------------
    async def inject_soakdb3_crystal_wells(
        self,
        visitid,
        models: List[Soakdb3CrystalWellModel],
        why="inject_soakdb3_crystal_wells",
    ) -> Dict:
        """
        Append the crystal wells described by the models
        into the soakdb3 database for the given visit.

        We don't insert the same CrystalPlate/CrystalWell twice.
        """

        self.__establish_soakdb3_dataface_client()

        # Get the necessary values from the (single) head table row.
        head_rows = await self.soakdb3_dataface_client.query_for_dictionary(  # type: ignore
            visitid,
            f"SELECT Protein, DropVolume FROM {Tablenames.HEAD}",
        )
        protein = head_rows[0]["Protein"]
        drop_volume = head_rows[0]["DropVolume"]

        logger.debug(describe("head row protein", protein))
        logger.debug(describe("head row drop_volume", drop_volume))

        # Get rows of all existing plate/well pairs in the soakdb3 database.
        plate_well_rows = await self.soakdb3_dataface_client.query(  # type: ignore
            visitid,
            f"SELECT ID, CrystalPlate, CrystalWell FROM {Tablenames.BODY} ORDER BY ID ASC",
        )

        # Flatten the plate_well values into a list of combine plate/well records.
        plate_wells = []
        blank_row_ids = []
        for plate_well_row in plate_well_rows:
            # This is a row with empty or blank CrystalPlate?
            if plate_well_row[1] is None or plate_well_row[1] == "":
                # Remember the ID of this row so we can update it.
                blank_row_ids.append(plate_well_row[0])
            else:
                plate_wells.append(
                    self.__plate_well(plate_well_row[1], plate_well_row[2]),
                )

        updated_count = 0
        inserted_count = 0
        skipped_count = 0

        # Loop over all the models to be appended.
        id_to_insert = 0
        fields = []
        for model in models:
            # Make combined plate/well name for this model.
            plate_well = self.__plate_well(
                model.CrystalPlate,
                model.CrystalWell,
            )
            # Already have this plate/well?
            if plate_well in plate_wells:
                skipped_count += 1
                continue
            plate_wells.append(plate_well)

            if len(blank_row_ids) == 0:
                # ID for this row is next negative number, causing insert.
                id_to_insert = id_to_insert - 1
                id = id_to_insert
                inserted_count += 1
            else:
                # ID for this row is one of the empty rows.
                id = int(blank_row_ids.pop(0))
                updated_count += 1
            # Make a row for each field in the model.
            record = model.dict()
            for field in list(record.keys()):
                # Ignore the ID from the model since it is of indeterminate value at this point.
                if field == "ID":
                    continue
                # Certain fields on the row come from the current head row field values.
                if field == "ProteinName":
                    value = protein
                elif field == "DropVolume":
                    value = drop_volume
                else:
                    value = record[field]
                fields.append(
                    {
                        "id": str(id),
                        "field": field,
                        "value": value,
                    }
                )

        await self.soakdb3_dataface_client.update_body_fields(  # type: ignore
            visitid,
            fields,
        )

        return {
            "updated_count": updated_count,
            "inserted_count": inserted_count,
            "skipped_count": skipped_count,
        }

    # ----------------------------------------------------------------------------------------
    async def fetch_soakdb3_crystal_wells_serialized(
        self,
        visitid: str,
        why=None,
    ) -> List[Dict]:
        """
        Caller provides the filters for selecting which crystal plates.
        Returns records from the database.

        TODO: Add a query filter to fetch_crystal_wells_serialized.
        """

        # Get the models from the direct call.
        models = await self.fetch_soakdb3_crystal_wells(visitid, why=why)

        # Serialize models into dicts to give to the response.
        records = [model.dict() for model in models]

        return records

    # ----------------------------------------------------------------------------------------
    async def fetch_soakdb3_crystal_wells(
        self,
        visitid: str,
        why: Optional[str] = None,
    ) -> List[Soakdb3CrystalWellModel]:
        """"""

        self.__establish_soakdb3_dataface_client()

        # Get rows of all existing plate/well pairs in the soakdb3 database.
        records = await self.soakdb3_dataface_client.query_for_dictionary(  # type: ignore
            visitid,
            f"SELECT * FROM {Tablenames.BODY} ORDER BY ID ASC",
        )

        # Dicts are returned, so parse them into models.
        # Fields which came from the query which are not defined in the model are ignored.
        # TODO: Fetch only desired fields when querying in fetch_soakdb3_crystal_wells.
        models = [Soakdb3CrystalWellModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    def __establish_soakdb3_dataface_client(self) -> None:
        """
        Get a soakdb3 dataface client client.

        Once a connection is made, the object reference is kept
        and the same one used as return for subsequent calls.
        """

        # TODO: Solve problem of DirectSoakdb3CrystalWells mixin constructor not getting called, and remove several # type: ignore.
        if not hasattr(self, "soakdb3_dataface_client"):
            self.soakdb3_dataface_client = None

        if self.soakdb3_dataface_client is not None:
            return

        soakdb3_specification = require(
            f"{callsign(self)} specification",
            self.specification(),
            "soakdb3_dataface_specification",
        )

        self.soakdb3_dataface_client = Soakdb3ApiDatafaces().build_object(
            soakdb3_specification
        )

    # ----------------------------------------------------------------------------------------
    def __plate_well(self, plate: str, well: str) -> str:
        """
        Make a combined string out of the plate and well pair.

        Args:
            plate (str): plate name
            well (str): well name

        Returns:
            str: plate/well combined
        """

        return f"{plate}.{well}"



from dls_utilpack.envvar import Envvar as DlsUtilpackEnvvar


class Envvar(DlsUtilpackEnvvar):
    """Class which covers environment variables, with default values."""

    XCHEMBKU_CONFIGFILE = "XCHEMBKU_CONFIGFILE"



import logging
from pathlib import Path

# Soakdb3 database.
from soakdb3_api.databases.constants import Tablenames

# Client for direct access to the soakdb3 database for seeding it.
from soakdb3_api.datafaces.context import Context as Soakdb3DatafaceClientContext
from soakdb3_api.datafaces.datafaces import (
    datafaces_get_default as soakdb3_datafaces_get_default,
)

# The model which describes the crystal wells to be injected into soakdb3.
from soakdb3_api.models.crystal_well_model import (
    CrystalWellModel as Soakdb3CrystalWellModel,
)

# The service process startup/teardown context.
from soakdb3_lib.datafaces.context import Context as Soakdb3DatafaceServerContext

# Base class for the tester.
from tests.base import Base

# Client context creator.
from xchembku_api.datafaces.context import Context as XchembkuDatafaceClientContext

# Object managing datafaces.
from xchembku_api.datafaces.datafaces import xchembku_datafaces_get_default

# Server context creator.
from xchembku_lib.datafaces.context import Context as XchembkuDatafaceServerContext

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestSoakdb3CrystalWellDirectSqlite:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_sqlite.yaml"
        Soakdb3CrystalWellTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestSoakdb3CrystalWellDirectMysql:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_mysql.yaml"
        Soakdb3CrystalWellTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestSoakdb3CrystalWellServiceSqlite:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_sqlite.yaml"
        Soakdb3CrystalWellTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestSoakdb3CrystalWellServiceMysql:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_mysql.yaml"
        Soakdb3CrystalWellTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class Soakdb3CrystalWellTester(Base):
    """
    Class to test the dataface droplocation-related endpoints.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """
        self.__injected_count = 0

        # Get the multiconf from the testing configuration yaml.
        multiconf = self.get_multiconf()

        # Load the multiconf into a dict.
        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the soakdb3 dataface.
        soakdb3_dataface_specification = multiconf_dict[
            "soakdb3_dataface_specification"
        ]

        # Make the soakdb3 server context.
        soakdb3_server_context = Soakdb3DatafaceServerContext(
            soakdb3_dataface_specification
        )

        # Make the soakdb3 CLIENT context.
        soakdb3_client_context = Soakdb3DatafaceClientContext(
            soakdb3_dataface_specification
        )

        # Reference the dict entry for the xchembku dataface.
        xchembku_dataface_specification = multiconf_dict[
            "xchembku_dataface_specification"
        ]

        # Make the xchembku server context.
        xchembku_server_context = XchembkuDatafaceServerContext(
            xchembku_dataface_specification
        )

        # Make the xchembku client context.
        xchembku_client_context = XchembkuDatafaceClientContext(
            xchembku_dataface_specification
        )

        # Start the soakdb3 server context which includes the direct or network-addressable service.
        async with soakdb3_server_context:
            # Client for direct access to the soakdb3 database for seeding it.
            async with soakdb3_client_context:
                # Start the xchembku server context which includes the direct or network-addressable service.
                async with xchembku_server_context:
                    # Start the matching xchembku client context.
                    async with xchembku_client_context:
                        await self.__run_the_test(constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_the_test(self, constants, output_directory):
        """ """

        # Reference the soakdb3 dataface object which the context has set up as the default.
        soakdb3_dataface = soakdb3_datafaces_get_default()

        # Reference the xchembku dataface object which the context has set up as the default.
        xchembku_dataface = xchembku_datafaces_get_default()

        models = []

        visit = "cm00001-1"
        visit_directory = Path(output_directory) / "visits" / visit
        visit_directory.mkdir(parents=True)

        # Soakdb3 expects visitid to be a visit directory.
        # This is because of how the soadkb3 VBA in the Excel works.
        visitid = str(visit_directory)

        seeded_crystal_plate = "98aa_2021-09-13_RI1000-0276-3drop"
        injected_crystal_plate = "98ab_2021-09-14_RI1000-0276-3drop"

        # ----------------------------------------------------------------
        # Seed the necessary fields in the head table.

        protein = "P1"
        drop_volume = 3.1

        head_record = {
            "Protein": protein,
            "DropVolume": drop_volume,
        }

        # Insert these fields as the (single) row in the soakdb3 database's head table.
        await soakdb3_dataface.insert(  # type: ignore
            visitid,
            Tablenames.HEAD,
            [head_record],
        )

        # ----------------------------------------------------------------
        seed_fields = []
        # Seed row ID 1 with no exising crystal well on it.
        # This row will absorb the first injected well.
        seed_fields.append(
            {
                "id": str(-1),
                "field": "ProteinName",
                "value": "something",
            }
        )
        # Seed row ID 2 with an exising crystal well on it.
        # This row will not be touched by injected wells.
        seed_fields.append(
            {
                "id": str(-2),
                "field": "CrystalPlate",
                "value": seeded_crystal_plate,
            }
        )

        # Send these seeds to the soakdb3 database's body table.
        await soakdb3_dataface.update_body_fields(  # type: ignore
            visitid,
            seed_fields,
        )

        # ----------------------------------------------------------------

        # Make some wells to insert.
        models.append(
            Soakdb3CrystalWellModel(
                LabVisit=visit,
                CrystalPlate=injected_crystal_plate,
                CrystalWell="01A1",
                EchoX=100,
                EchoY=200,
            )
        )

        # Same well again, should be ignored.
        models.append(
            Soakdb3CrystalWellModel(
                LabVisit=visit,
                CrystalPlate=injected_crystal_plate,
                CrystalWell="01A1",
                EchoX=101,
                EchoY=201,
            )
        )

        # Different well.
        models.append(
            Soakdb3CrystalWellModel(
                LabVisit=visit,
                CrystalPlate=injected_crystal_plate,
                CrystalWell="01A2",
                EchoX=200,
                EchoY=300,
            )
        )

        # Write crystal well records.
        await xchembku_dataface.inject_soakdb3_crystal_wells(visitid, models)

        # Check the results
        queried_models = await xchembku_dataface.fetch_soakdb3_crystal_wells(visitid)
        assert len(queried_models) == 3

        # Make sure the original location is not overwritten.
        assert queried_models[0].EchoX == 100

        # A new different well.
        models.append(
            Soakdb3CrystalWellModel(
                LabVisit=visit,
                CrystalPlate=injected_crystal_plate,
                CrystalWell="01A3",
                EchoX=300,
                EchoY=400,
            )
        )

        # Write the full list of crystal well records again
        models[0].EchoX = 103
        await xchembku_dataface.inject_soakdb3_crystal_wells(visitid, models)

        # Check the results, there should be no change to the first ones.
        queried_models = await xchembku_dataface.fetch_soakdb3_crystal_wells(visitid)
        assert len(queried_models) == 4
        assert queried_models[0].ID == "1"
        assert queried_models[0].CrystalWell == "01A1"
        assert queried_models[0].CrystalPlate == injected_crystal_plate
        assert queried_models[0].ProteinName == protein
        assert queried_models[0].DropVolume == drop_volume

        # Row ID 2 was seeded and this don't get the injected crystal well.
        assert queried_models[1].ID == "2"
        assert queried_models[1].CrystalPlate == seeded_crystal_plate
        assert queried_models[1].ProteinName is None
        assert queried_models[1].DropVolume is None

        # The last two should have been injected.
        assert queried_models[2].ID == "3"
        assert queried_models[2].CrystalWell == "01A2"
        assert queried_models[2].CrystalPlate == injected_crystal_plate
        assert queried_models[2].ProteinName == protein
        assert queried_models[2].DropVolume == drop_volume

        assert queried_models[3].ID == "4"
        assert queried_models[3].CrystalWell == "01A3"
        assert queried_models[3].CrystalPlate == injected_crystal_plate
        assert queried_models[3].ProteinName == protein
        assert queried_models[3].DropVolume == drop_volume

        # Make sure the original location is not overwritten.
        assert queried_models[0].EchoX == 100



import copy
import logging
from typing import Any, Dict, List

from dls_normsql.constants import CommonFieldnames

from xchembku_api.models.crystal_well_filter_model import (
    CrystalWellFilterModel,
    CrystalWellFilterSortbyEnum,
)
from xchembku_api.models.crystal_well_model import CrystalWellModel
from xchembku_api.models.crystal_well_needing_droplocation_model import (
    CrystalWellNeedingDroplocationModel,
)
from xchembku_lib.datafaces.direct_base import DirectBase

logger = logging.getLogger(__name__)


class DirectCrystalWells(DirectBase):
    """ """

    # ----------------------------------------------------------------------------------------
    async def upsert_crystal_wells_serialized(
        self,
        records: List[Dict],
        why=None,
    ) -> Dict:
        # We are being given json, so parse it into models.
        models = [CrystalWellModel(**record) for record in records]
        # Return the method doing the work.
        return await self.upsert_crystal_wells(models, why=why)

    # ----------------------------------------------------------------------------------------
    async def upsert_crystal_wells(
        self,
        models: List[CrystalWellModel],
        why="upsert_crystal_wells",
    ) -> Dict:
        """
        Caller provides the crystal well record with the fields to be updated.

        We don't insert the same filename twice.

        TODO: Consider an alternate way besides filename to distinguish duplicate crystal wells in upsert.

        TODO: Find more efficient way to upsert_crystal_wells in batch.
        """

        inserted_count = 0
        updated_count = 0

        # Loop over all the models to be upserted.
        for model in models:
            # Find any existing record for this model object.
            records = await self.query(
                "SELECT * FROM crystal_wells WHERE filename = ?",
                subs=[model.filename],
                why=why,
            )

            if len(records) > 0:
                # Make a copy of the model record and remove some fields not to update.
                model_copy = copy.deepcopy(model.dict())
                model_copy.pop(CommonFieldnames.UUID)
                model_copy.pop(CommonFieldnames.CREATED_ON)
                model_copy.pop("filename")
                model_copy.pop("crystal_plate_uuid")
                result = await self.update(
                    "crystal_wells",
                    model_copy,
                    "(filename = ?)",
                    subs=[model.filename],
                    why=why,
                )
                updated_count += result.get("count", 0)
            else:
                await self.insert(
                    "crystal_wells",
                    [model.dict()],
                    why=why,
                )
                inserted_count += 1

        return {
            "updated_count": updated_count,
            "inserted_count": inserted_count,
        }

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_wells_filenames_serialized(
        self, limit: int = 1, why=None
    ) -> List[Dict]:
        """ """

        # Get the models from the direct call.
        models = await self.fetch_crystal_wells_filenames(limit=limit, why=why)

        # Serialize models into dicts to give to the response.
        records = [model.dict() for model in models]

        return records

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_wells_filenames(
        self, limit: int = 1, why=None
    ) -> List[CrystalWellModel]:
        """
        Filenams for ALL wells ever.
        """

        if why is None:
            why = "API fetch_crystal_wells_filenames"
        records = await self.query(
            "SELECT"
            " crystal_wells.uuid,"
            " crystal_wells.position,"
            " crystal_wells.filename,"
            " crystal_wells.crystal_plate_uuid,"
            f" crystal_wells.{CommonFieldnames.CREATED_ON}"
            f" FROM crystal_wells"
            f" ORDER BY {CommonFieldnames.CREATED_ON}",
            why=why,
        )

        # Parse the records returned by sql into models.
        models = [CrystalWellModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_wells_needing_autolocation_serialized(
        self, limit: int = 1, why=None
    ) -> List[Dict]:
        """ """

        # Get the models from the direct call.
        models = await self.fetch_crystal_wells_needing_autolocation(
            limit=limit, why=why
        )

        # Serialize models into dicts to give to the response.
        records = [model.dict() for model in models]

        return records

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_wells_needing_autolocation(
        self, limit: int = 1, why=None
    ) -> List[CrystalWellModel]:
        """
        Wells need an autolocation if they don't have one yet.
        """

        if why is None:
            why = "API fetch_crystal_wells_needing_autolocation"
        records = await self.query(
            "SELECT crystal_wells.*"
            f"\n  FROM crystal_wells"
            f"\n  LEFT JOIN crystal_well_autolocations"
            " ON crystal_wells.uuid = crystal_well_autolocations.crystal_well_uuid"
            "\n  WHERE crystal_well_autolocations.uuid IS NULL"
            f"\n  ORDER BY {CommonFieldnames.CREATED_ON}"
            f"\n  LIMIT {limit}",
            why=why,
        )

        # Parse the records returned by sql into models.
        models = [CrystalWellModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_wells_needing_droplocation_serialized(
        self, filter: Dict, why=None
    ) -> List[Dict]:
        """
        Caller provides the filters for selecting which crystal wells.
        Returns records from the database.
        """

        # Get the models from the direct call.
        models = await self.fetch_crystal_wells_needing_droplocation(
            CrystalWellFilterModel(**filter), why=why
        )

        # Serialize models into dicts to give to the response.
        records = [model.dict() for model in models]

        return records

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_wells_needing_droplocation(
        self, filter: CrystalWellFilterModel, why=None
    ) -> List[CrystalWellNeedingDroplocationModel]:
        """
        Wells need a droplocation if they have an autolocation.
        """

        # Caller wants results relative to anchor?
        if filter.anchor is not None and filter.direction is not None:
            return await self.__fetch_crystal_wells_needing_droplocation_hard(
                filter, why=why
            )
        # Query can be made easier if there is no anchor with direction involved.
        else:
            return await self.__fetch_crystal_wells_needing_droplocation_easy(
                filter, why=why
            )

    # ----------------------------------------------------------------------------------------
    async def __fetch_crystal_wells_needing_droplocation_easy(
        self, filter: CrystalWellFilterModel, why=None
    ) -> List[CrystalWellNeedingDroplocationModel]:
        """
        Wells need a droplocation if they have an autolocation.
        """

        if why is None:
            why = "API fetch_crystal_wells_needing_droplocation"

        # Build the individual pieces of the SQL query.
        subs: List[Any] = []
        orderby = self.__build_orderby(filter)
        where = self.__build_where(filter, subs)
        fields = self.__build_fields(filter)
        joins = self.__build_joins(filter)

        # Glue them together.
        main_query = "\nSELECT" + fields + joins + where + "\n" + orderby

        if filter.limit is not None:
            main_query += f"\nLIMIT {filter.limit}"

        # Query the database.
        records = await self.query(main_query, subs=subs, why=why)

        # Parse the records returned by sql into models.
        models = [CrystalWellNeedingDroplocationModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def __fetch_crystal_wells_needing_droplocation_hard(
        self, filter: CrystalWellFilterModel, why=None
    ) -> List[CrystalWellNeedingDroplocationModel]:
        """
        This is the query when we want records relative from an anchor record.

        Since this involves complex subqueries, we limit use of this
        to when we're looking only at records in a specific visit or plate.
        """

        if filter.visit is None:
            raise RuntimeError(
                "programming error: no visit supplied with relative crystal well filter"
            )

        if why is None:
            why = "API fetch_crystal_wells_needing_droplocation"

        # Build the individual pieces of the SQL query.
        subs: List[Any] = []
        orderby = self.__build_orderby(filter)
        where = self.__build_where(filter, subs)
        fields = self.__build_fields(filter)
        joins = self.__build_joins(filter)

        # We need the row number to be in both the main and sub query.
        row_number = f"\n  ROW_NUMBER() OVER ({orderby}) AS ordered_row_number"
        fields = row_number + "," + fields

        # Main query gets the actual results.
        main_query = "SELECT" + fields + joins + where

        # Build another "where" since it may add subs.
        where = self.__build_where(filter, subs)

        # Sub query computes row_numbers under the same filter in the same order as the main query.
        sub_query = "\nSELECT\n  crystal_wells.uuid," + row_number + joins + where

        # Do the main query, but filter the results by the subquery based on matching row numbers.
        full_query = (
            f"\nSELECT * FROM (\n{main_query}\n) AS main_query"
            f"\n/* Match row_numbers starting from the anchor {filter.anchor}. */"
            "\nWHERE ordered_row_number >"
            f"\n  (SELECT ordered_row_number FROM ({sub_query}\n) AS sub_query"
            f"\n    WHERE sub_query.uuid = ?)"
            f"\n    ORDER BY ordered_row_number"
        )
        subs.append(filter.anchor)

        if filter.limit is not None:
            full_query += f"\nLIMIT {filter.limit}"

        # Do the actual query.
        records = await self.query(full_query, subs=subs, why=why)

        # Parse the records returned by sql into models.
        models = [CrystalWellNeedingDroplocationModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    def __build_fields(
        self,
        filter: CrystalWellFilterModel,
    ) -> str:
        """
        Wells need a droplocation if they have an autolocation.
        """

        fields = (
            "\n  crystal_wells.*,"
            "\n  crystal_well_autolocations.auto_target_x,"
            "\n  crystal_well_autolocations.auto_target_y,"
            "\n  crystal_well_autolocations.well_centroid_x,"
            "\n  crystal_well_autolocations.well_centroid_y,"
            "\n  crystal_well_autolocations.drop_detected,"
            "\n  crystal_well_autolocations.number_of_crystals,"
            "\n  crystal_well_droplocations.confirmed_target_x,"
            "\n  crystal_well_droplocations.confirmed_target_y,"
            "\n  crystal_well_droplocations.confirmed_microns_x,"
            "\n  crystal_well_droplocations.confirmed_microns_y,"
            "\n  crystal_well_droplocations.is_usable,"
            "\n  crystal_well_droplocations.is_exported_to_soakdb3,"
            "\n  crystal_plates.visit,"
            "\n  crystal_plates.thing_type AS crystal_plate_thing_type,"
            "\n  crystal_plates.rockminer_collected_stem"
        )

        return fields

    # ----------------------------------------------------------------------------------------
    def __build_joins(
        self,
        filter: CrystalWellFilterModel,
    ) -> str:
        """
        Wells need a droplocation if they have an autolocation.
        """

        joins = (
            "\nFROM crystal_wells"
            "\nJOIN crystal_well_autolocations ON crystal_well_autolocations.crystal_well_uuid = crystal_wells.uuid"
            "\nLEFT JOIN crystal_well_droplocations ON crystal_well_droplocations.crystal_well_uuid = crystal_wells.uuid"
            "\nLEFT JOIN crystal_plates ON crystal_plates.uuid = crystal_wells.crystal_plate_uuid"
        )

        return joins

    # ----------------------------------------------------------------------------------------
    def __build_where(
        self,
        filter: CrystalWellFilterModel,
        subs: List[Any],
    ) -> str:
        """
        Wells need a droplocation if they have an autolocation.
        """

        where = "WHERE"
        sql = ""

        # Caller wants a glob of file?
        if filter.filename_pattern is not None:
            sql += (
                "\n/* Just certain filenames. */"
                f"\n{where} crystal_wells.filename REGEXP ?"
            )
            subs.append(filter.filename_pattern)
            where = "AND"

        # Caller wants specific barcode?
        if filter.barcode is not None:
            sql += (
                f"\n/* Just wells on plate with barcode '{filter.barcode}'. */"
                f"\n{where} crystal_plates.barcode = ?"
            )
            subs.append(filter.barcode)
            where = "AND"

        # Caller wants specific visit?
        if filter.visit is not None:
            sql += (
                f"\n/* Just wells on plates with visit '{filter.visit}'. */"
                f"\n{where} crystal_plates.visit = ?"
            )
            subs.append(filter.visit)
            where = "AND"

        # Caller wants only those not yet decided?
        if filter.is_decided is False:
            sql += (
                "\n/* Include only crystal wells which have not had a decision made. */"
                f"\n{where} crystal_well_droplocations.is_usable IS NULL"
            )
            where = "AND"

        # Caller wants only those which are decided?
        # Confirmed means a droplocation record has been created at all (though might not have usable coordinates).
        if filter.is_decided is True:
            sql += (
                "\n/* Include only crystal wells which have a decision made. */"
                f"\n{where} crystal_well_droplocations.is_usable IS NOT NULL"
            )
            where = "AND"

        # Caller wants only those which are decided but do or don't have usable coordinates?
        if filter.is_usable is not None:
            sql += (
                f"\n/* Include only crystal wells which have filter.is_usable = {filter.is_usable}. */"
                f"\n{where} crystal_well_droplocations.is_usable = ?"
            )
            subs.append(filter.is_usable)
            where = "AND"

        # Caller wants only those which are exported to soakdb3?
        if filter.is_exported_to_soakdb3 is not None:
            sql += (
                f"\n/* Include only crystal wells which have been exported to soakdb3. */"
                f"\n{where} crystal_well_droplocations.is_usable = True"
            )
            where = "AND"

        # Caller wants just the anchor record?
        if filter.anchor is not None and filter.direction is None:
            sql += (
                f"\n/* Get the crystal well at the anchor {filter.anchor}. */"
                f"\n{where} crystal_wells.uuid = ?"
            )
            subs.append(filter.anchor)
            where = "AND"

        # Caller wants those in direction from the anchor record?
        # This means we ne need the anchor to be in the rows,
        # no matter if it would have been excluded by other filters.
        if filter.anchor is not None and filter.direction is not None:
            # We presume that all previous conjunctions were AND
            # which have higher operator precedence than OR,
            # so we can stick an OR on here at the end
            if where == "AND":
                where = "OR"
            sql += (
                "\n/* Always include the crystal well at the anchor. */"
                f"\n{where} crystal_wells.uuid = ?"
            )
            subs.append(filter.anchor)
            where = "AND"

        return sql

    # ----------------------------------------------------------------------------------------
    def __build_orderby(
        self,
        filter: CrystalWellFilterModel,
    ) -> str:

        sql = ""

        position_direction = "ASC"
        if filter.direction == -1:
            position_direction = "DESC"

        # Filter says order by number of crystals?
        if filter.sortby == CrystalWellFilterSortbyEnum.NUMBER_OF_CRYSTALS:
            crystals_direction = "DESC"
            if filter.direction == -1:
                crystals_direction = "ASC"

            # If duplicate crystals, use position as tie breaker.
            order_by = f"crystal_well_autolocations.number_of_crystals {crystals_direction}, crystal_wells.position {position_direction}"
        else:
            order_by = f"crystal_wells.position {position_direction}"

        sql += f"ORDER BY {order_by}"

        return sql



import logging
from typing import Dict, List, Optional

from soakdb3_api.models.crystal_well_model import (
    CrystalWellModel as Soakdb3CrystalWellModel,
)

# Class for an aiohttp client.
from xchembku_api.aiohttp_client import AiohttpClient

# Dataface protocolj things.
from xchembku_api.datafaces.constants import Commands, Keywords
from xchembku_api.models.crystal_plate_filter_model import CrystalPlateFilterModel
from xchembku_api.models.crystal_plate_model import CrystalPlateModel
from xchembku_api.models.crystal_plate_report_model import CrystalPlateReportModel
from xchembku_api.models.crystal_well_autolocation_model import (
    CrystalWellAutolocationModel,
)
from xchembku_api.models.crystal_well_droplocation_model import (
    CrystalWellDroplocationModel,
)
from xchembku_api.models.crystal_well_filter_model import CrystalWellFilterModel
from xchembku_api.models.crystal_well_model import CrystalWellModel
from xchembku_api.models.crystal_well_needing_droplocation_model import (
    CrystalWellNeedingDroplocationModel,
)

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------------------
class Aiohttp(AiohttpClient):
    """
    Object implementing client side API for talking to the xchembku_dataface server.
    Please see doctopic [A01].
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):

        # We will get an umbrella specification which must contain an aiohttp_specification within it.
        AiohttpClient.__init__(
            self,
            specification["type_specific_tbd"]["aiohttp_specification"],
        )

    # ----------------------------------------------------------------------------------------
    async def query(self, sql, subs=None, why: Optional[str] = None):
        """"""
        return await self.__send_protocolj(
            "query",
            sql,
            subs=subs,
            why=why,
        )

    # ----------------------------------------------------------------------------------------
    async def execute(self, sql, subs=None, why: Optional[str] = None):
        """"""
        return await self.__send_protocolj(
            "execute",
            sql,
            subs=subs,
            why=why,
            as_transaction=True,
        )

    # ----------------------------------------------------------------------------------------
    async def insert(self, table_name, records, why: Optional[str] = None):
        """"""
        return await self.__send_protocolj(
            "insert",
            table_name,
            records,
            why=why,
            as_transaction=True,
        )

    # ----------------------------------------------------------------------------------------
    async def update(
        self, table_name, record, where, subs=None, why: Optional[str] = None
    ):
        """"""
        return await self.__send_protocolj(
            "update",
            table_name,
            record,
            where,
            subs=subs,
            why=why,
            as_transaction=True,
        )

    # ----------------------------------------------------------------------------------------
    async def upsert_crystal_plates(
        self,
        models: List[CrystalPlateModel],
        why: Optional[str] = None,
    ) -> None:
        """"""

        records: List[Dict] = [model.dict() for model in models]
        await self.__send_protocolj(
            "upsert_crystal_plates_serialized",
            records,
            why=why,
            as_transaction=True,
        )

        return None

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_plates(
        self,
        filter: CrystalPlateFilterModel,
        why: Optional[str] = None,
    ) -> List[CrystalPlateModel]:
        """"""

        records = await self.__send_protocolj(
            "fetch_crystal_plates_serialized",
            filter=filter.dict(),
            why=why,
        )

        # Dicts are returned, so parse them into models.
        models = [CrystalPlateModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def report_crystal_plates(
        self,
        filter: CrystalPlateFilterModel,
        why: Optional[str] = None,
    ) -> List[CrystalPlateReportModel]:
        """"""

        records = await self.__send_protocolj(
            "report_crystal_plates_serialized",
            filter=filter.dict(),
            why=why,
        )

        # Dicts are returned, so parse them into models.
        models = [CrystalPlateReportModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_wells_filenames(
        self,
        limit: int = 1,
        why: Optional[str] = None,
    ) -> List[CrystalWellModel]:
        """"""

        records = await self.__send_protocolj(
            "fetch_crystal_wells_filenames_serialized",
            limit=limit,
            why=why,
        )

        # Dicts are returned, so parse them into models.
        models = [CrystalWellModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def upsert_crystal_wells(
        self,
        models: List[CrystalWellModel],
    ) -> None:
        """"""

        records: List[Dict] = [model.dict() for model in models]
        await self.__send_protocolj(
            "upsert_crystal_wells_serialized",
            records,
            as_transaction=True,
        )

        return None

    # ----------------------------------------------------------------------------------------
    async def update_crystal_wells(
        self,
        records,
        why: Optional[str] = None,
        as_transaction=True,
    ) -> Dict:
        """"""

        return await self.__send_protocolj(
            "update_crystal_wells_serialized",
            records,
            why=why,
        )

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_wells_needing_autolocation(
        self,
        limit: int = 1,
        why: Optional[str] = None,
    ) -> List[CrystalWellModel]:
        """"""

        records = await self.__send_protocolj(
            "fetch_crystal_wells_needing_autolocation_serialized",
            limit=limit,
            why=why,
        )

        # Dicts are returned, so parse them into models.
        models = [CrystalWellModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def fetch_crystal_wells_needing_droplocation(
        self,
        filter: CrystalWellFilterModel,
        why: Optional[str] = None,
    ) -> List[CrystalWellNeedingDroplocationModel]:
        """"""

        records = await self.__send_protocolj(
            "fetch_crystal_wells_needing_droplocation_serialized",
            filter.dict(),
            why=why,
        )

        # Dicts are returned, so parse them into models.
        models = [CrystalWellNeedingDroplocationModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def originate_crystal_well_autolocations(
        self, models: List[CrystalWellAutolocationModel]
    ) -> None:
        """"""

        records: List[Dict] = [model.dict() for model in models]
        await self.__send_protocolj(
            "originate_crystal_well_autolocations_serialized",
            records,
            as_transaction=True,
        )

        return None

    # ----------------------------------------------------------------------------------------
    async def upsert_crystal_well_droplocations(
        self,
        models: List[CrystalWellDroplocationModel],
        only_fields: Optional[List[str]] = None,
        why: Optional[str] = None,
    ) -> None:
        """"""

        records: List[Dict] = [model.dict() for model in models]
        await self.__send_protocolj(
            "upsert_crystal_well_droplocations_serialized",
            records,
            only_fields=only_fields,
            why=why,
            as_transaction=True,
        )

        return None

    # ----------------------------------------------------------------------------------------
    async def inject_soakdb3_crystal_wells(
        self,
        visitid: str,
        models: List[Soakdb3CrystalWellModel],
        why: Optional[str] = None,
    ) -> Dict:
        """"""

        records: List[Dict] = [model.dict() for model in models]
        result = await self.__send_protocolj(
            "inject_soakdb3_crystal_wells_serialized",
            visitid,
            records,
            why=why,
            as_transaction=True,
        )
        return result

    # ----------------------------------------------------------------------------------------
    async def fetch_soakdb3_crystal_wells(
        self,
        visitid: str,
        why: Optional[str] = None,
    ) -> List[Soakdb3CrystalWellModel]:
        """"""

        records = await self.__send_protocolj(
            "fetch_soakdb3_crystal_wells_serialized",
            visitid,
            why=why,
        )

        # Dicts are returned, so parse them into models.
        models = [Soakdb3CrystalWellModel(**record) for record in records]

        return models

    # ----------------------------------------------------------------------------------------
    async def __send_protocolj(self, function, *args, **kwargs):
        """"""

        return await self.client_protocolj(
            {
                Keywords.COMMAND: Commands.EXECUTE,
                Keywords.PAYLOAD: {
                    "function": function,
                    "args": args,
                    "kwargs": kwargs,
                },
            },
        )



import logging
from typing import Optional

# Base class for the tester.
from tests.base import Base

# Types which the CrystalPlateObjects factory can use to build an instance.
from xchembku_api.crystal_plate_objects.constants import (
    ThingTypes as CrystalPlateThingTypes,
)

# Client context creator.
from xchembku_api.datafaces.context import Context as XchembkuDatafaceClientContext

# Object managing datafaces.
from xchembku_api.datafaces.datafaces import xchembku_datafaces_get_default
from xchembku_api.models.crystal_plate_model import CrystalPlateModel
from xchembku_api.models.crystal_well_autolocation_model import (
    CrystalWellAutolocationModel,
)
from xchembku_api.models.crystal_well_droplocation_model import (
    CrystalWellDroplocationModel,
)
from xchembku_api.models.crystal_well_filter_model import CrystalWellFilterModel
from xchembku_api.models.crystal_well_model import CrystalWellModel
from xchembku_lib.crystal_plate_objects.crystal_plate_objects import CrystalPlateObjects

# Server context creator.
from xchembku_lib.datafaces.context import Context as XchembkuDatafaceServerContext

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestCrystalWellDroplocation1DirectSqlite:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_sqlite.yaml"
        CrystalWellDroplocation1Tester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class TestCrystalWellDroplocation1DirectMysql:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_mysql.yaml"
        CrystalWellDroplocation1Tester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class TestCrystalWellDroplocation1ServiceSqlite:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_sqlite.yaml"
        CrystalWellDroplocation1Tester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class TestCrystalWellDroplocation1ServiceMysql:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_mysql.yaml"
        CrystalWellDroplocation1Tester().main(
            constants, configuration_file, output_directory
        )


# ----------------------------------------------------------------------------------------
class CrystalWellDroplocation1Tester(Base):
    """
    Class to test the dataface droplocation-related endpoints.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """
        self.__injected_count = 0

        # Get the multiconf from the testing configuration yaml.
        multiconf = self.get_multiconf()

        # Load the multiconf into a dict.
        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the xchembku dataface.
        xchembku_dataface_specification = multiconf_dict[
            "xchembku_dataface_specification"
        ]

        # Make the server context.
        xchembku_server_context = XchembkuDatafaceServerContext(
            xchembku_dataface_specification
        )

        # Make the client context.
        xchembku_client_context = XchembkuDatafaceClientContext(
            xchembku_dataface_specification
        )

        # Start the xchembku server context which includes the direct or network-addressable service.
        async with xchembku_server_context:
            # Start the matching xchembku client context.
            async with xchembku_client_context:
                await self.__run_the_test(constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_the_test(self, constants, output_directory):
        """ """

        # Reference the dataface object which the context has set up as the default.
        dataface = xchembku_datafaces_get_default()

        # Make a plate for the wells we will create.
        self.__visit = "cm00001-1"
        self.__barcode = "xyzw"
        self.__crystal_plate_model = CrystalPlateModel(
            formulatrix__plate__id=1,
            barcode=self.__barcode,
            visit=self.__visit,
            thing_type=CrystalPlateThingTypes.SWISS3,
        )

        self.__crystal_plate_object = CrystalPlateObjects().build_object(
            {"type": CrystalPlateThingTypes.SWISS3}
        )

        # Write plate record.
        await dataface.upsert_crystal_plates(
            [self.__crystal_plate_model],
        )

        models = []

        # Inject some wells, all of which will belong to the plate we just made.
        models.append(await self.__inject(dataface, False, False))
        models.append(await self.__inject(dataface, True, True))
        models.append(await self.__inject(dataface, True, False))
        models.append(await self.__inject(dataface, True, True))
        models.append(await self.__inject(dataface, True, True))
        models.append(await self.__inject(dataface, True, False))

        # --------------------------------------------------------------------------
        # Query for list from filename glob.
        crystal_well_models = await self.__check(
            dataface,
            CrystalWellFilterModel(filename_pattern=".*A_1.jpg"),
            3,
            "filename glob",
            filename="02A_1.jpg",
        )

        assert (
            crystal_well_models[1].filename == "03A_1.jpg"
        ), "filename glob, second response"
        assert (
            crystal_well_models[2].filename == "04A_1.jpg"
        ), "filename glob, third response"

        # --------------------------------------------------------------------------

        # Check the filtered queries.
        await self.__check(dataface, CrystalWellFilterModel(), 5, "no limit, all")

        # Upsert all with the same values to make sure it doesn't make new records.
        records = await dataface.query(
            "SELECT * FROM crystal_well_droplocations",
            why="[UPSCHK] direct query all drop locations",
        )
        m = [CrystalWellDroplocationModel(**record) for record in records]
        await dataface.upsert_crystal_well_droplocations(m, why="[UPSCHK]")
        await self.__check(
            dataface, CrystalWellFilterModel(), 5, "no limit, all after upsert"
        )

        await self.__check(dataface, CrystalWellFilterModel(limit=1), 1, "limit 1")
        await self.__check(dataface, CrystalWellFilterModel(limit=2), 2, "limit 2")
        await self.__check(
            dataface, CrystalWellFilterModel(is_decided=True), 3, "confirmed only"
        )
        await self.__check(
            dataface,
            CrystalWellFilterModel(barcode=self.__barcode, is_decided=True),
            3,
            "confirmed only, barcode",
        )
        await self.__check(
            dataface,
            CrystalWellFilterModel(barcode="abcd", is_decided=True),
            0,
            "confirmed only, other barcode",
        )
        await self.__check(
            dataface, CrystalWellFilterModel(is_decided=False), 2, "unconfirmed only"
        )

        # Check the anchor query forward.
        await self.__check(
            dataface,
            CrystalWellFilterModel(
                visit=self.__visit,
                anchor=models[3].uuid,
                direction=1,
                limit=1,
            ),
            1,
            "anchored forward",
            filename="05B_1.jpg",
        )

        # Check the anchor query forward at the end of the list.
        await self.__check(
            dataface,
            CrystalWellFilterModel(
                visit=self.__visit,
                anchor=models[5].uuid,
                direction=1,
            ),
            0,
            "anchored forward at the end of the list",
        )

        # Check the anchor query backward.
        await self.__check(
            dataface,
            CrystalWellFilterModel(
                visit=self.__visit,
                anchor=models[2].uuid,
                direction=-1,
                limit=1,
            ),
            1,
            "anchored backward",
            filename="02A_1.jpg",
        )

        # Check the anchor query backward at the start of the list.
        await self.__check(
            dataface,
            CrystalWellFilterModel(
                visit=self.__visit,
                anchor=models[1].uuid,
                direction=-1,
            ),
            0,
            "anchored at the start of the list",
        )

        # Check the anchor query backward at the start of the list of those unconfirmed.
        await self.__check(
            dataface,
            CrystalWellFilterModel(
                is_decided=False,
                visit=self.__visit,
                anchor=models[1].uuid,
                direction=-1,
            ),
            0,
            "anchored at the start of the list, backward, unconfirmed",
        )

        # Check the anchor query backward at the start of the list of those unconfirmed.
        await self.__check(
            dataface,
            CrystalWellFilterModel(
                is_decided=False,
                visit=self.__visit,
                anchor=models[2].uuid,
                direction=-1,
            ),
            0,
            "anchored at the start of the list, forward unconfirmed",
        )

        # --------------------------------------------------------------------------
        # Check the usable queries.
        await self.__check(
            dataface,
            CrystalWellFilterModel(is_decided=True, is_usable=False),
            0,
            "confirmed but unusable only (1)",
        )
        crystal_well_models = await self.__check(
            dataface,
            CrystalWellFilterModel(
                visit=self.__visit,
                is_usable=True,
            ),
            3,
            "usable only, visit",
        )

        crystal_well_models = await self.__check(
            dataface,
            CrystalWellFilterModel(
                barcode=self.__barcode,
                is_usable=True,
            ),
            3,
            "usable only, barcode",
        )

        # Change one of the usable to unusable.
        t = CrystalWellDroplocationModel(
            crystal_well_uuid=crystal_well_models[0].uuid,
            is_usable=False,
        )

        await dataface.upsert_crystal_well_droplocations([t])

        # Check the usable queries again.
        await self.__check(
            dataface,
            CrystalWellFilterModel(
                visit=self.__visit,
                is_decided=True,
                is_usable=False,
            ),
            1,
            "confirmed but unusable only (2)",
        )
        crystal_well_models = await self.__check(
            dataface,
            CrystalWellFilterModel(is_usable=True),
            2,
            "usable only after upsert",
        )

    # ----------------------------------------------------------------------------------------

    async def __inject(self, dataface, autolocation: bool, droplocation: bool):
        """ """

        letter = "A"
        if self.__injected_count > 3:
            letter = "B"

        self.__injected_count += 1
        filename = "%02d%s_1.jpg" % (self.__injected_count, letter)
        position = "%s%02da" % (letter, self.__injected_count)

        # Create the well object.
        m = CrystalWellModel(
            position=position,
            crystal_plate_uuid=self.__crystal_plate_model.uuid,
            crystal_plate_thing_type=self.__crystal_plate_model.thing_type,
            filename=filename,
        )

        # Write well record.
        await dataface.upsert_crystal_wells([m])

        if autolocation:
            # Add a crystal well autolocation.
            ta = CrystalWellAutolocationModel(
                crystal_well_uuid=m.uuid,
                number_of_crystals=self.__injected_count,
                well_centroid_x=100,
                well_centroid_y=100,
            )

            await dataface.originate_crystal_well_autolocations([ta])

        if droplocation:
            # Add a crystal well droplocation.
            td = CrystalWellDroplocationModel(
                crystal_well_uuid=m.uuid,
                confirmed_target_x=150,
                confirmed_target_y=50,
                is_usable=True,
            )

            await dataface.upsert_crystal_well_droplocations([td])

        return m

    # ----------------------------------------------------------------------------------------

    async def __check(
        self,
        dataface,
        filter: CrystalWellFilterModel,
        expected: int,
        note: str,
        filename: Optional[str] = None,
    ):
        """ """

        # Get the full crystal well with auto and confirmed drop locations.
        crystal_well_models = await dataface.fetch_crystal_wells_needing_droplocation(
            filter
        )

        # Make sure we got enough.
        assert len(crystal_well_models) == expected, note

        if filename is not None:
            assert crystal_well_models[0].filename == filename, f"{note} filename"

        for crystal_well_model in crystal_well_models:
            # All wells should belong to the visit.
            assert crystal_well_model.visit == self.__visit, f"{note} visit"

            # All present confirmed drop locations should be consistent.
            if (
                crystal_well_model.well_centroid_x is not None
                and crystal_well_model.confirmed_target_x is not None
                and crystal_well_model.well_centroid_y is not None
                and crystal_well_model.confirmed_target_y is not None
            ):

                # Use the template plate to predict the microns.
                (
                    microns_x,
                    microns_y,
                ) = self.__crystal_plate_object.compute_drop_location_microns(
                    crystal_well_model.dict()
                )
                assert crystal_well_model.confirmed_target_x == 150
                assert crystal_well_model.confirmed_microns_x == microns_x
                assert crystal_well_model.confirmed_target_y == 50
                assert crystal_well_model.confirmed_microns_y == microns_y

        return crystal_well_models



import logging

# Base class.
from dls_utilpack.client_context_base import ClientContextBase

# Things created in the context.
from xchembku_api.datafaces.datafaces import Datafaces, xchembku_datafaces_set_default

logger = logging.getLogger(__name__)


class Context(ClientContextBase):
    """
    Client context for a xchembku_dataface object.
    On entering, it creates the object according to the specification (a dict).
    On exiting, it closes client connection.

    The aenter and aexit methods are exposed for use by an enclosing context and the base class.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        ClientContextBase.__init__(self, specification)

    # ----------------------------------------------------------------------------------------
    async def aenter(self):
        """ """

        # Build the object according to the specification.
        self.interface = Datafaces().build_object(self.specification)

        # If there is more than one dataface, the last one defined will be the default.
        xchembku_datafaces_set_default(self.interface)

        # Open client session to the service or direct connection.
        await self.interface.open_client_session()

        # For convenience, return the object which is the client interface.
        return self.interface

    # ----------------------------------------------------------------------------------------
    async def aexit(self):
        """ """
        if self.interface is not None:
            # Close client session to the service or direct connection.
            await self.interface.close_client_session()

            # Clear the global variable.  Important between pytests.
            xchembku_datafaces_set_default(None)



import uuid
from typing import List, Optional, Tuple

from pydantic import BaseModel


class CrystalWellAutolocationModel(BaseModel):
    """
    Model containing crystal well autolocation information.

    Typically this structure is populated and transmitted by the chimpflow.
    """

    uuid: str
    crystal_well_uuid: Optional[str] = None
    drop_detected: Optional[bool] = None
    auto_target_x: Optional[int] = None
    auto_target_y: Optional[int] = None
    well_centroid_x: Optional[int] = None
    well_centroid_y: Optional[int] = None
    number_of_crystals: Optional[int] = None
    crystal_coordinates: Optional[List[Tuple[int, int]]] = None

    # TODO: Add proper pydantic date parsing/valiation to CREATED_ON fields.
    created_on: Optional[str] = None

    def __init__(self, **kwargs):
        # Automatically cook up a uuid if it's not provided to the constructor.
        if "uuid" not in kwargs:
            kwargs["uuid"] = str(uuid.uuid4())
        super().__init__(**kwargs)

    # For temporary reference, this is what xchem_chimp was doing to send to echolocator database:
    # request_dict[ImageFieldnames.FILENAME] = str(im_path)
    # if output_dict["drop_detected"] is True:
    #     request_dict[ImageFieldnames.IS_DROP] = True
    #     echo_y, echo_x = output_dict["echo_coordinate"][0]
    #     request_dict[ImageFieldnames.TARGET_POSITION_Y] = int(echo_y)
    #     request_dict[ImageFieldnames.TARGET_POSITION_X] = int(echo_x)
    #     centroid_y, centroid_x = output_dict["well_centroid"]
    #     request_dict[ImageFieldnames.WELL_CENTER_Y] = int(centroid_y)
    #     request_dict[ImageFieldnames.WELL_CENTER_X] = int(centroid_x)
    #     num_xtals = len(output_dict["xtal_coordinates"])
    #     request_dict[ImageFieldnames.NUMBER_OF_CRYSTALS] = int(num_xtals)
    #     logging.info(f"output_dict is\n{describe('output_dict', output_dict)}")
    # else:
    #     request_dict[ImageFieldnames.IS_DROP] = False
    #     request_dict[ImageFieldnames.IS_USABLE] = False
    # logging.info(f"Sending request for {im_path} to EchoLocator database")
    # asyncio.run(self.send_item_to_echolocator(request_dict))



class Keywords:
    COMMAND = "xchembku_datafaces::keywords::command"
    PAYLOAD = "xchembku_datafaces::keywords::payload"


class Commands:
    EXECUTE = "xchembku_datafaces::commands::execute"


class Types:
    AIOHTTP = "xchembku_lib.xchembku_datafaces.aiohttp"
    DIRECT = "xchembku_lib.xchembku_datafaces.direct"



from typing import Optional

from pydantic import BaseModel


class CrystalWellNeedingDroplocationModel(BaseModel):
    """
    Model containing well information formed from a composite query of droplocation information.

    Typically this structure is returned by queries.
    """

    # Stuff from the original record.
    uuid: str
    crystal_plate_uuid: str
    position: str
    filename: str
    width: Optional[int]
    height: Optional[int]
    error: Optional[str]
    created_on: str

    # Stuff from the plate.
    visit: str
    crystal_plate_thing_type: str
    rockminer_collected_stem: Optional[str] = None

    # Stuff from the autolocation.
    auto_target_x: Optional[int] = None
    auto_target_y: Optional[int] = None
    well_centroid_x: Optional[int] = None
    well_centroid_y: Optional[int] = None
    drop_detected: Optional[bool] = None
    number_of_crystals: Optional[int] = None

    # Stuff from the droplocation.
    crystal_well_droplocation_uuid: Optional[str] = None
    confirmed_target_x: Optional[int] = None
    confirmed_target_y: Optional[int] = None
    confirmed_microns_x: Optional[int] = None
    confirmed_microns_y: Optional[int] = None

    is_usable: Optional[bool] = None
    is_exported_to_soakdb3: Optional[bool] = None



import logging
from typing import Optional

# Base class for the tester.
from tests.base import Base

# Types which the CrystalPlateObjects factory can use to build an instance.
from xchembku_api.crystal_plate_objects.constants import (
    ThingTypes as CrystalPlateObjectThingTypes,
)

# Client context creator.
from xchembku_api.datafaces.context import Context as XchembkuDatafaceClientContext

# Object managing datafaces.
from xchembku_api.datafaces.datafaces import xchembku_datafaces_get_default
from xchembku_api.models.crystal_plate_filter_model import CrystalPlateFilterModel
from xchembku_api.models.crystal_plate_model import CrystalPlateModel

# Crystal plate objects factory.
from xchembku_lib.crystal_plate_objects.crystal_plate_objects import CrystalPlateObjects

# Server context creator.
from xchembku_lib.datafaces.context import Context as XchembkuDatafaceServerContext

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestCrystalPlateDirectSqlite:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_sqlite.yaml"
        CrystalPlateTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestCrystalPlateDirectMysql:
    """
    Test dataface interface by direct call.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        configuration_file = "tests/configurations/direct_mysql.yaml"
        CrystalPlateTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestCrystalPlateServiceSqlite:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_sqlite.yaml"
        CrystalPlateTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestCrystalPlateServiceMysql:
    """
    Test dataface interface through network interface.
    """

    def test(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """ """

        configuration_file = "tests/configurations/service_mysql.yaml"
        CrystalPlateTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class CrystalPlateTester(Base):
    """
    Class to test the dataface droplocation-related endpoints.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """
        # Get the multiconf from the testing configuration yaml.
        multiconf = self.get_multiconf()

        # Load the multiconf into a dict.
        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the xchembku dataface.
        xchembku_dataface_specification = multiconf_dict[
            "xchembku_dataface_specification"
        ]

        # Make the server context.
        xchembku_server_context = XchembkuDatafaceServerContext(
            xchembku_dataface_specification
        )

        # Make the client context.
        xchembku_client_context = XchembkuDatafaceClientContext(
            xchembku_dataface_specification
        )

        # Start the xchembku server context which includes the direct or network-addressable service.
        async with xchembku_server_context:
            # Start the matching xchembku client context.
            async with xchembku_client_context:
                await self.__run_the_test(constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_the_test(self, constants, output_directory):
        """ """

        # Reference the dataface object which the context has set up as the default.
        dataface = xchembku_datafaces_get_default()

        visit = "cm00001-1"
        models = []
        models.append(
            CrystalPlateModel(
                formulatrix__plate__id=10,
                barcode="xyz1",
                visit=visit,
            )
        )
        models.append(
            CrystalPlateModel(
                formulatrix__plate__id=20,
                barcode="xyz2",
                visit=visit,
            )
        )
        models.append(
            CrystalPlateModel(
                formulatrix__plate__id=30,
                barcode="xyz3",
                visit=visit,
            )
        )

        await dataface.upsert_crystal_plates(models)

        # Check the filtered queries.
        await self.__check(
            dataface,
            CrystalPlateFilterModel(),
            3,
            "all",
        )
        await self.__check(
            dataface,
            CrystalPlateFilterModel(limit=2),
            2,
            "limit",
        )
        await self.__check(
            dataface,
            CrystalPlateFilterModel(limit=1, direction=1),
            1,
            "earliest",
            formulatrix__plate__id=10,
        )
        await self.__check(
            dataface,
            CrystalPlateFilterModel(limit=1, direction=-1),
            1,
            "latest",
            formulatrix__plate__id=30,
        )
        await self.__check(
            dataface,
            CrystalPlateFilterModel(uuid=models[1].uuid),
            1,
            "by uuid",
            formulatrix__plate__id=20,
        )
        await self.__check(
            dataface,
            CrystalPlateFilterModel(barcode=models[2].barcode),
            1,
            "by barcode",
            formulatrix__plate__id=30,
        )
        await self.__check(
            dataface,
            CrystalPlateFilterModel(
                from_formulatrix__plate__id=models[0].formulatrix__plate__id
            ),
            2,
            "from plate_id",
            formulatrix__plate__id=20,
        )

        # ------------------------------------------------------------------------------------
        # Create an object to be upserted.
        upserted_model = CrystalPlateModel(
            formulatrix__plate__id=40,
            barcode="xyz4",
            visit=visit,
            thing_type=CrystalPlateObjectThingTypes.SWISS3,
        )

        # First upsert is an insert.
        await dataface.upsert_crystal_plates([upserted_model])
        upserted_models = await self.__check(
            dataface,
            CrystalPlateFilterModel(limit=1, direction=-1),
            1,
            "upsert insert",
            formulatrix__plate__id=40,
        )
        assert upserted_models[0].uuid == upserted_model.uuid
        created_on = upserted_models[0].created_on

        # Again, with no change, so should not insert.
        await dataface.upsert_crystal_plates([upserted_model])
        upserted_models = await self.__check(
            dataface,
            CrystalPlateFilterModel(limit=1, direction=-1),
            1,
            "upsert insert",
            formulatrix__plate__id=40,
        )
        assert upserted_models[0].uuid == upserted_model.uuid
        assert upserted_models[0].created_on == created_on

        # Again, this time changing visit.
        upserted_model.visit = visit + "1"
        await dataface.upsert_crystal_plates([upserted_model])
        upserted_models = await self.__check(
            dataface,
            CrystalPlateFilterModel(limit=1, direction=-1),
            1,
            "upsert insert",
            formulatrix__plate__id=40,
        )

        assert upserted_models[0].uuid == upserted_model.uuid
        assert upserted_models[0].created_on == created_on
        assert upserted_models[0].visit == upserted_model.visit

        # Make sure that the model which came out of the database can be instantiated.
        specification = {"type": upserted_models[0].thing_type}
        crystal_plate_model = CrystalPlateObjects().build_object(specification)
        assert crystal_plate_model.get_well_count() == 288

        # TODO: Move testing of drop location in microns out of test_crystal_plate.py and into its own test file.
        # Cook up a fake crystal well so we can use the plate to convert drop location to microns.
        crystal_well_dict = {
            "well_centroid_x": 100,
            "well_centroid_y": 101,
            "confirmed_target_x": 150,
            "confirmed_target_y": 51,
        }

        # Let the plate decide how to convert pixels into microns.
        x_microns, y_microns = crystal_plate_model.compute_drop_location_microns(
            crystal_well_dict
        )
        assert x_microns == int(0.5 + 2.837 * 50)
        assert y_microns == int(0.5 + 2.837 * -50)

    # ----------------------------------------------------------------------------------------

    async def __check(
        self,
        dataface,
        filter: CrystalPlateFilterModel,
        expected: int,
        note: str,
        formulatrix__plate__id: Optional[int] = None,
    ):
        """ """

        crystal_plate_models = await dataface.fetch_crystal_plates(filter)

        assert len(crystal_plate_models) == expected, note

        if formulatrix__plate__id is not None:
            assert (
                crystal_plate_models[0].formulatrix__plate__id == formulatrix__plate__id
            ), f"{note} formulatrix__plate__id"

        return crystal_plate_models



import logging
from typing import Dict

# Database manager.
from dls_normsql.databases import Databases
from dls_utilpack.callsign import callsign

# Base class for generic things.
from dls_utilpack.thing import Thing

from xchembku_api.databases.database_definition import DatabaseDefinition

logger = logging.getLogger(__name__)

thing_type = "xchembku_lib.xchembku_datafaces.direct"


class DirectBase(Thing):
    """ """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):
        Thing.__init__(self, thing_type, specification)

        # For testing, caller might want to drop the database on connection.
        self.__should_drop_database = specification.get("should_drop_database")

        self.__database_definition_object = DatabaseDefinition()

        self.__database = None

    # ----------------------------------------------------------------------------------------
    async def start(self):
        # Connect to the database to create the schemas if they don't exist already.
        await self.establish_database_connection()

        # Make sure we are up to date with the latest database schema revision.
        await self.__database.apply_revisions()

    # ----------------------------------------------------------------------------------------
    async def disconnect(self):
        if self.__database is not None:
            logger.debug(f"[DISSHU] {callsign(self)} disconnecting")
            await self.__database.disconnect()
            logger.debug(f"[DISSHU] {callsign(self)} disconnected")
            self.__database = None

        # TODO: Figure out a better way to disconnect the dataface mixins.
        await self.disconnect_soakdb3_crystal_wells_mixin()

    # ----------------------------------------------------------------------------------------
    async def establish_database_connection(self):
        if self.__database is None:
            self.__database = Databases().build_object(
                self.specification()["database"],
                self.__database_definition_object,
            )

            # For testing, caller might want to drop the database on connection.
            await self.__database.connect(
                should_drop_database=self.__should_drop_database
            )

    # ----------------------------------------------------------------------------------------
    async def reinstance(self):
        """"""
        if self.__database is None:
            return

        self.__database = self.__database.reinstance()

        return self

    # ----------------------------------------------------------------------------------------
    async def backup(self):
        """"""
        await self.establish_database_connection()

        return await self.__database.backup()

    # ----------------------------------------------------------------------------------------
    async def restore(self, nth):
        """"""
        await self.establish_database_connection()

        return await self.__database.restore(nth)

    # ----------------------------------------------------------------------------------------
    async def query(self, sql, subs=None, why=None):
        """"""

        await self.establish_database_connection()

        records = await self.__database.query(sql, subs=subs, why=why)

        return records

    # ----------------------------------------------------------------------------------------
    async def execute(self, sql, subs=None, why=None):
        """"""
        await self.establish_database_connection()

        return await self.__database.execute(sql, subs=subs, why=why)

    # ----------------------------------------------------------------------------------------
    async def insert(self, table_name, records, why=None) -> None:
        """"""
        await self.establish_database_connection()

        return await self.__database.insert(table_name, records, why=why)

    # ----------------------------------------------------------------------------------------
    async def update(self, table_name, record, where, subs=None, why=None) -> Dict:
        """"""
        await self.establish_database_connection()

        if why is None:
            why = f"update {table_name} record"

        # This returns the count of records changed by the update.
        return {
            "count": await self.__database.update(
                table_name, record, where, subs=subs, why=why
            )
        }

    # ----------------------------------------------------------------------------------------
    async def begin(self, why=None) -> None:
        """"""
        await self.establish_database_connection()

        return await self.__database.begin()

    # ----------------------------------------------------------------------------------------
    async def commit(self, why=None) -> None:
        """"""
        await self.establish_database_connection()

        return await self.__database.commit()

    # ----------------------------------------------------------------------------------------
    async def rollback(self, why=None) -> None:
        """"""
        await self.establish_database_connection()

        return await self.__database.rollback()

    # ----------------------------------------------------------------------------------------
    async def report_health(self):
        """"""

        report = {}

        report["alive"] = True

        return report

    # ----------------------------------------------------------------------------------------
    async def open_client_session(self):
        """"""
        # Connect to the database to create the schemas if they don't exist already.
        await self.establish_database_connection()

    # ----------------------------------------------------------------------------------------
    async def close_client_session(self):
        """"""
        logger.debug(f"[DISSHU] {callsign(self)} in aexit, calling disconnect")

        await self.disconnect()

        logger.debug(f"[DISSHU] {callsign(self)} in aexit, disconnected")



import logging
from typing import Dict, List

from xchembku_api.models.crystal_well_autolocation_model import (
    CrystalWellAutolocationModel,
)
from xchembku_lib.datafaces.direct_base import DirectBase

logger = logging.getLogger(__name__)


class DirectCrystalWellAutolocations(DirectBase):
    """ """

    # ----------------------------------------------------------------------------------------
    async def originate_crystal_well_autolocations_serialized(
        self, records: List[Dict]
    ) -> None:
        # We are being given json, so parse it into models.
        models = [CrystalWellAutolocationModel(**record) for record in records]
        # Return the method doing the work.
        return await self.originate_crystal_well_autolocations(models)

    # ----------------------------------------------------------------------------------------
    async def originate_crystal_well_autolocations(
        self, models: List[CrystalWellAutolocationModel]
    ) -> None:
        """
        Caller provides the records containing fields to be created.
        """

        # We're being given models, serialize them into dicts for the sql.
        records = [model.dict() for model in models]

        return await self.insert(
            "crystal_well_autolocations",
            records,
            why="originate_crystal_well_autolocations",
        )



import uuid
from typing import Optional

from pydantic import BaseModel


class CrystalWellDroplocationModel(BaseModel):
    """
    Model containing crystal well droplocation information.

    Typically this structure is populated and transmitted by the chimpflow.
    """

    uuid: str
    crystal_well_uuid: Optional[str] = None
    confirmed_target_x: Optional[int] = None
    confirmed_target_y: Optional[int] = None
    confirmed_microns_x: Optional[int] = None
    confirmed_microns_y: Optional[int] = None
    is_usable: Optional[bool] = None

    # True when exported to soakdb3.
    is_exported_to_soakdb3: Optional[bool] = None

    # TODO: Add proper pydantic date parsing/valiation to CREATED_ON fields.
    created_on: Optional[str] = None

    def __init__(self, **kwargs):
        # Automatically cook up a uuid if it's not provided to the constructor.
        if "uuid" not in kwargs:
            kwargs["uuid"] = str(uuid.uuid4())
        super().__init__(**kwargs)



import logging

# Base class for an asyncio server context.
from dls_utilpack.server_context_base import ServerContextBase

# Things created in the context.
from xchembku_lib.datafaces.datafaces import Datafaces

logger = logging.getLogger(__name__)


thing_type = "xchembku_lib.xchembku_datafaces.context"


class Context(ServerContextBase):
    """
    Asyncio context for a xchembku_dataface server object.
    On entering, it creates the object according to the specification (a dict).
    If configured, it starts the server as a coroutine, thread or process.
    On exiting, it commands the server to shut down.

    The enter and exit methods are exposed for use during testing.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        ServerContextBase.__init__(self, thing_type, specification)

    # ----------------------------------------------------------------------------------------
    async def aenter(self):
        """ """

        # Build the object according to the specification.
        self.server = Datafaces().build_object(self.specification())

        # If there is more than one collector, the last one defined will be the default.
        # collectors_set_default(self.server)

        if self.context_specification.get("start_as") == "coro":
            await self.server.activate_coro()

        elif self.context_specification.get("start_as") == "thread":
            await self.server.start_thread()

        elif self.context_specification.get("start_as") == "process":
            await self.server.start_process()

    # ----------------------------------------------------------------------------------------
    async def aexit(self, type=None, value=None, traceback=None):
        """
        Asyncio context exit.

        Stop service if one was started and releases any client resources.
        """
        logger.debug(f"[DISSHU] {thing_type} aexit")

        if self.server is not None:
            if self.context_specification.get("start_as") == "process":
                # The server associated with this context is running?
                if await self.is_process_alive():
                    logger.debug(f"[DISSHU] {thing_type} calling client_shutdown")
                    # Put in request to shutdown the server.
                    await self.server.client_shutdown()

            if self.context_specification.get("start_as") == "coro":
                await self.server.direct_shutdown()



import logging

import dls_mainiac_lib.version

import xchembku_lib.version

from . import __version__

logger = logging.getLogger(__name__)


# ----------------------------------------------------------
def version():
    """
    Current version.
    """

    return __version__


# ----------------------------------------------------------
def meta(given_meta=None):
    """
    Returns version information as a dict.
    Adds version information to given meta, if any.
    """
    s = {}
    s["xchembku_cli"] = version()
    s.update(xchembku_lib.version.meta())
    s.update(dls_mainiac_lib.version.meta())

    if given_meta is not None:
        given_meta.update(s)
    else:
        given_meta = s
    return given_meta
